import numpy as np
import random

# Latent heat of vaporisation [MJ.Kg-1]
LAMBDA = 2.45

#: Solar constant [ MJ m-2 min-1]
SOLAR_CONSTANT = 0.0820

SB_CONS = 3.405e-12  # per minute


ALLOWED_COLUMNS = ['temp', 'tmin', 'tmax', 'rel_hum', 'sol_rad', 'rn', 'wind_speed', 'es', 'sunshine_hrs',
                   'rh_min', 'rh_max', 'rh_mean', 'et_rad', 'cs_rad', 'half_hr', 'tdew', 'daylight_hrs']


def random_array(_length, lower=0.0, upper=0.5):
    """This creates a random array of length `length` and the floats vary between `lower` and `upper`."""
    rand_array = np.zeros(_length)
    for i in range(_length):
        rand_array[i] = random.uniform(lower, upper)
    return rand_array


def Colors(_name):

    colors = {'sol_rad': np.array([0.84067393, 0.66066663, 0.22888342]),
              'rel_hum': np.array([0.50832319, 0.53790088, 0.17337983]),
              'rh_mean': np.array([0.34068018, 0.65708722, 0.19501699]),
              'tmax': np.array([0.94943837, 0.34234137, 0.03188675]),
              'tmin': np.array([0.91051202, 0.65414968, 0.48220781]),
              'ra': np.array([0.64027147, 0.75848599, 0.59123481]),
              'rn': np.array([0.68802968, 0.38316639, 0.13177745]),
              'ea': np.array([0.19854081, 0.44556471, 0.35620562]),
              'wind_speed': np.array([0.40293934, 0.51160837, 0.1293387]),
              'jday': np.array([0.88396109, 0.14081036, 0.44402598]),
              'half_hr': np.array([0.53974518, 0.48519598, 0.11808065]),
              'sha': np.array([0.67873225, 0.67178641, 0.01953063])}

    if 'Daily' in _name:
        return np.array([0.63797563, 0.05503074, 0.07078517])
    elif 'Minute' in _name:
        return np.array([0.27822191, 0.7608274, 0.89536561])
    elif 'Hourly' in _name:
        return np.array([0.70670405, 0.71039014, 0.54375619])
    elif 'Monthly' in _name:
        return np.array([0.39865179, 0.61455622, 0.57515074])
    elif 'Yearly' in _name:
        return np.array([0.81158386, 0.182704, 0.93272506])
    elif 'Sub_hourly' in _name:
        return np.array([0.1844271, 0.70936978, 0.53026012])
    elif _name in colors:
        return colors[_name]
    else:
        c = random_array(3, 0.01, 0.99)
        print('for ', _name, c)
        return c


default_constants = {
    'lat': {'desc': 'latitude in decimal degrees', 'def_val': None, 'min': -90, 'max': 90},
    'long': {'desc': 'longitude in decimal degrees', 'def_val': None},
    'a_s': {'desc': 'fraction of extraterrestrial radiation reaching earth on sunless days', 'def_val': 0.23},
    'b_s': {'desc': """difference between fracion of extraterrestrial radiation reaching full-sun days and that 
                    on sunless days""", 'def_val': 0.5},
    'albedo': {'desc': """a numeric value between 0 and 1 (dimensionless), albedo of evaporative surface representing
     the portion of the incident radiation that is reflected back at the surface. Default is 0.23 for
    surface covered with short reference crop, which is for the calculation of Matt-Shuttleworth
     reference crop evaporation.""", 'def_val': 0.23, 'min': 0, 'max': 1},
    'abtew_k': {'desc': 'a coefficient used in Abtew', 'def_val': 0.52},
    'CH': {'desc': 'crop height', 'def_val': 0.12},
    'Ca': {'desc': 'Specific heat of air', 'def_val': 0.001013},
    'surf_res': {'desc': "surface resistance (s/m) depends on the type of reference crop. Default is 70 for short"
                         " reference crop", 'def_val': 70, 'min': 0, 'max': 9999},
    'alphaPT': {'desc': 'Priestley-Taylor coefficient', 'def_val': 1.26},
    'ritchie_a': {'desc': "Coefficient for Richie method", 'def_val': 0.002322},
    'ritchie_b': {'desc': "Coefficient for Richie method", 'def_val': 0.001548},
    'ritchie_c': {'desc': "Coefficient for Ritchie Method", 'def_val': 0.11223}
}



"""

This code is almost exact copy and past of keract library except few changes
in _get_nodes function where I added one if else statement The original code at
https://github.com/philipperemy/keract has following MIT licence as of 24 Nov 2020.

MIT License

Copyright (c) 2019 Philippe Rémy

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from collections import OrderedDict

from .backend import tf, keras

if keras is not None:
    K = keras.backend
    Sequential = tf.keras.Sequential
    Model = tf.keras.models.Model
else:
    K, Sequential, Model = None, None, None


def is_placeholder(n):
    return (hasattr(n, '_op') and n._op.type == 'Placeholder') or '_input' in str(n) or 'input' in str(n)

def n_(node, output_format, nested=False, module=None):
    if isinstance(node, list):
        node_name = '_'.join([str(n.name) for n in node])
    else:
        node_name = str(node.name)
    if module is not None and nested:
        node_name = module.name + '/' + node_name
    if output_format == 'simple' and ':' in node_name:
        return node_name.split(':')[0]
    elif output_format == 'full' and hasattr(node, 'output'):
        return node.output.name
    return node_name


def _evaluate(model: Model, nodes_to_evaluate, x, y=None, auto_compile=False):

    if not model._is_compiled:
        if auto_compile:
            model.compile(loss='mse', optimizer='adam')
        else:
            raise Exception('Compilation of the model required.')

    def eval_fn(k_inputs):
        try:
            return K.function(k_inputs, nodes_to_evaluate)(model._standardize_user_data(x, y))
        except AttributeError:  # one way to avoid forcing non eager mode.
            if y is None:  # tf 2.3.0 upgrade compatibility.
                return K.function(k_inputs, nodes_to_evaluate)(x)
            return K.function(k_inputs, nodes_to_evaluate)((x, y))  # although works.
        except ValueError as e:
            print('Run it without eager mode tf.compat.v1.disable_eager_execution()')
            raise e

    try:
        return eval_fn(model._feed_inputs + model._feed_targets + model._feed_sample_weights)
    except Exception:
        return eval_fn(model._feed_inputs)


def get_gradients_of_trainable_weights(model, x, y):
    """
    Get the gradients of trainable_weights for the kernel and the bias nodes for all filters in each layer.
    Trainable_weights gradients are averaged over the minibatch.
    :param model: keras compiled model
    :param x: inputs for which gradients are sought (averaged over all inputs if batch_size > 1)
    :param y: outputs for which gradients are sought
    :return: dict mapping layers to corresponding gradients (filter_h, filter_w, in_channels, out_channels)
    """
    nodes = OrderedDict([(n.name, n) for n in model.trainable_weights])
    return _get_gradients(model, x, y, nodes)


def get_gradients_of_activations(model, x, y, layer_names=None, output_format='simple', nested=False):
    """
    Get gradients of the outputs of the activation functions, regarding the loss.
    Intuitively, it shows how your activation maps change over a tiny modification of the loss.
    :param model: keras compiled model or one of ['vgg16', 'vgg19', 'inception_v3', 'inception_resnet_v2',
    'mobilenet_v2', 'mobilenetv2'].
    :param x: Model input (Numpy array). In the case of multi-inputs, x should be of type List.
    :param y: Model target (Numpy array). In the case of multi-inputs, y should be of type List.
    :param layer_names: (optional) Single name of a layer or list of layer names for which activations should be
    returned. It is useful in very big networks when it is computationally expensive to evaluate all the layers/nodes.
    :param output_format: Change the output dictionary key of the function.
    - 'simple': output key will match the names of the Keras layers. For example Dense(1, name='d1') will
    return {'d1': ...}.
    - 'full': output key will match the full name of the output layer name. In the example above, it will
    return {'d1/BiasAdd:0': ...}.
    - 'numbered': output key will be an index range, based on the order of definition of each layer within the model.
    :param nested: (optional) If set, will move recursively through the model definition to retrieve nested layers.
                Recursion ends at leaf layers of the model tree or at layers with their name specified in layer_names.

                E.g., a model with the following structure

                -layer1
                    -conv1
                    ...
                    -fc1
                -layer2
                    -fc2

                ... yields a dictionary with keys 'layer1/conv1', ..., 'layer1/fc1', 'layer2/fc2'.
                If layer_names = ['layer2/fc2'] is specified, the dictionary will only hold one key 'layer2/fc2'.

                The layer names are generated by joining all layers from top level to leaf level with the separator '/'.
    :return: Dict {layer_names (specified by output_format) -> activation of the layer output/node (Numpy array)}.
    """
    nodes = OrderedDict()
    _get_nodes(model, nodes, output_format, nested=nested, layer_names=layer_names)
    return _get_gradients(model, x, y, nodes)


def _get_gradients(model, x, y, nodes):
    if model.optimizer is None:
        raise Exception('Please compile the model first. The loss function is required to compute the gradients.')

    nodes_names = nodes.keys()
    nodes_values = nodes.values()

    try:
        if not hasattr(model, 'total_loss'):
            raise Exception('Disable TF eager mode to use get_gradients.\n'
                            'Add this command at the beginning of your script:\n'
                            'tf.compat.v1.disable_eager_execution()')
        grads = model.optimizer.get_gradients(model.total_loss, nodes_values)
    except ValueError as e:
        if 'differentiable' in str(e):
            # Probably one of the gradients operations is not differentiable...
            grads = []
            differentiable_nodes = []
            for n in nodes_values:
                try:
                    grads.extend(model.optimizer.get_gradients(model.total_loss, n))
                    differentiable_nodes.append(n)
                except ValueError:
                    pass
            # nodes_values = differentiable_nodes
        else:
            raise e

    gradients_values = _evaluate(model, grads, x, y)

    return OrderedDict(zip(nodes_names, gradients_values))


def _get_nodes(module, nodes, output_format, nested=False, layer_names=None, depth=0):
    def update_node(n):
        is_node_a_model = isinstance(n, (Model, Sequential))
        if not is_placeholder(n):
            if is_node_a_model and nested:
                return
            try:
                mod = None if depth == 0 else module
                name = n_(n, output_format, nested, mod)
                if layer_names is None or name in layer_names:
                    if is_node_a_model:
                        if hasattr(n, '_layers'):
                            output = n._layers[-1].output
                        else:
                            output = n.layers[-1].output
                    else:
                        output = n.output
                    nodes.update({name: output})
            except AttributeError:
                pass

    try:
        layers = module._layers if hasattr(module, '_layers') else module.layers
    except AttributeError:
        return
    for layer in layers:
        update_node(layer)
        if nested:
            _get_nodes(layer, nodes, output_format, nested, layer_names, depth + 1)


def get_activations(model, x, layer_names=None, nodes_to_evaluate=None,
                    output_format='simple', nested=False, auto_compile=True):
    """
    Fetch activations (nodes/layers outputs as Numpy arrays) for a Keras model and an input X.
    By default, all the activations for all the layers are returned.
    :param model: Keras compiled model or one of ['vgg16', 'vgg19', 'inception_v3', 'inception_resnet_v2',
    'mobilenet_v2', 'mobilenetv2', ...].
    :param x: Model input (Numpy array). In the case of multi-inputs, x should be of type List.
    :param layer_names: (optional) Single name of a layer or list of layer names for which activations should be
    returned. It is useful in very big networks when it is computationally expensive to evaluate all the layers/nodes.
    :param nodes_to_evaluate: (optional) List of Keras nodes to be evaluated. Useful when the nodes are not
    in model.layers.
    :param output_format: Change the output dictionary key of the function.
    - 'simple': output key will match the names of the Keras layers. For example Dense(1, name='d1') will
    return {'d1': ...}.
    - 'full': output key will match the full name of the output layer name. In the example above, it will
    return {'d1/BiasAdd:0': ...}.
    - 'numbered': output key will be an index range, based on the order of definition of each layer within the model.
    :param nested: If specified, will move recursively through the model definition to retrieve nested layers.
                Recursion ends at leaf layers of the model tree or at layers with their name specified in layer_names.

                E.g., a model with the following structure

                -layer1
                    -conv1
                    ...
                    -fc1
                -layer2
                    -fc2

                ... yields a dictionary with keys 'layer1/conv1', ..., 'layer1/fc1', 'layer2/fc2'.
                If layer_names = ['layer2/fc2'] is specified, the dictionary will only hold one key 'layer2/fc2'.

                The layer names are generated by joining all layers from top level to leaf level with the separator '/'.
    :param auto_compile: If set to True, will auto-compile the model if needed.
    :return: Dict {layer_name (specified by output_format) -> activation of the layer output/node (Numpy array)}.
    """
    layer_names = [layer_names] if isinstance(layer_names, str) else layer_names
    # print('Layer names:', layer_names)
    nodes = OrderedDict()
    if nodes_to_evaluate is None:
        _get_nodes(model, nodes, output_format, nested, layer_names, auto_compile)
    else:
        if layer_names is not None:
            raise ValueError('Do not specify a [layer_name] with [nodes_to_evaluate]. It will not be used.')
        nodes = OrderedDict([(n_(node, 'full'), node) for node in nodes_to_evaluate])

    if len(nodes) == 0:
        if layer_names is not None:
            network_layers = ', '.join([layer.name for layer in model.layers])
            raise KeyError('Could not find a layer with name: [{}]. '
                           'Network layers are [{}]'.format(', '.join(layer_names), network_layers))
        else:
            raise ValueError('Nodes list is empty. Or maybe the model is empty.')

    # The placeholders are processed later (Inputs node in Keras). Due to a small bug in tensorflow.
    input_layer_outputs = []
    layer_outputs = OrderedDict()

    for key, node in nodes.items():
        if isinstance(node, list):
            for nod in node:
                if not is_placeholder(nod):
                    if key not in layer_outputs:
                        layer_outputs[key] = []
                    layer_outputs[key].append(nod)
        else:
            if not is_placeholder(node):
                layer_outputs.update({key: node})
    if nodes_to_evaluate is None or (layer_names is not None) and \
            any([n.name in layer_names for n in model.inputs]):
        input_layer_outputs = list(model.inputs)

    if len(layer_outputs) > 0:
        activations = _evaluate(model, layer_outputs.values(), x, y=None)
    else:
        activations = {}

    def craft_output(output_format_):
        inputs = [x] if not isinstance(x, list) else x
        activations_inputs_dict = OrderedDict(
            zip([n_(output, output_format_) for output in input_layer_outputs], inputs))
        activations_dict = OrderedDict(zip(layer_outputs.keys(), activations))
        result_ = activations_inputs_dict.copy()
        result_.update(activations_dict)

        if output_format_ == 'numbered':
            result_ = OrderedDict([(i, v) for i, (k, v) in enumerate(result_.items())])
        return result_

    result = craft_output(output_format)
    if layer_names is not None:  # extra check.
        result = {k: v for k, v in result.items() if k in layer_names}
    if nodes_to_evaluate is not None and len(result) != len(nodes_to_evaluate):
        result = craft_output(output_format_='full')  # collision detected in the keys.

    return result




__all__ = ["DLRegressionExperiments"]

import os

from ._main import Experiments
from ai4water import Model
from ai4water.hyperopt import Integer, Real, Categorical
from ai4water.utils.utils import jsonize, dateandtime_now


class DLRegressionExperiments(Experiments):
    """
    a framework for comparing several basic DL architectures for a given data.

    Any method defining a model should have the following signature:
    def model_<name>(self, **kwargs)
        self.param_space = ...
        self.x0 = ...
        layers = update_layers(**kwargs) # update layers with kwargs
        return self._make_return(layers, **kwargs)
    
    To check the available models 
    >>> exp = DLRegressionExperiments(...)
    >>> exp.models

    This class can be extended to different output shapes by overwriting 
    `:py:meth:ai4water.experiments.DLExperiments._output_layer` method.

    If learning rate, batch size, and lookback are are to be optimzied,
    their space can be specified in the following way:
    >>> exp = DLRegressionExperiments(...)
    >>> exp.lookback_space = [Integer(1, 100, name='lookback')]

    Example
    -------
    >>> from ai4water.experiments import DLRegressionExperiments
    >>> from ai4water.datasets import busan_beach
    >>> data = busan_beach()
    >>> exp = DLRegressionExperiments(
    >>> input_features = data.columns.tolist()[0:-1],
    >>> output_features = data.columns.tolist()[-1:],
    >>> epochs=300,
    >>> train_fraction=1.0,
    >>> y_transformation="log",
    >>> x_transformation="minmax",
    >>> )

    >>> exp.fit(data=data)

    """
    def __init__(self,
        param_space=None,
        x0=None,
        cases: dict = None,
        exp_name: str = None,
        num_samples: int = 5,
        verbosity: int = 1,
        **model_kws):

        self.param_space = param_space
        self.x0 = x0
        self.model_kws = model_kws

        self.lookback_space = []
        self.batch_size_space = []
        self.lr_space = []

        exp_name = exp_name or 'DLExperiments' + f'_{dateandtime_now()}'

        super().__init__(cases=cases,
                         exp_name=exp_name,
                         num_samples=num_samples,
                         verbosity=verbosity)

    @property
    def lookback_space(self):
        return self._lookback_space
    
    @lookback_space.setter
    def lookback_space(self, space):
        self._lookback_space = space

    @property
    def batch_size_space(self):
        return self._batch_size_space
    
    @batch_size_space.setter
    def batch_size_space(self, bs_space):
        self._batch_size_space = bs_space

    @property
    def lr_space(self):
        return self._lr_space
    
    @lr_space.setter
    def lr_space(self, lr_space):
        self._lr_space = lr_space

    @property
    def static_space(self):
        _space = []
        if self.lookback_space:
            _space.append(self.lookback_space)
        if self.batch_size_space:
            _space.append(self.batch_size_space)
        if self.lr_space:
            _space.append(self.lr_space)
        return _space

    @property
    def static_x0(self):
        _x0 = []
        if self.lookback_space:
            _x0.append(self.model_kws.get('lookback', 8))
        if self.batch_size_space:
            _x0.append(self.model_kws.get('batch_size', 32))
        if self.lr_space:
            _x0.append(self.model_kws.get('lr', 0.001))
        return _x0

    @property
    def mode(self):
        return "regression"

    @property
    def tpot_estimator(self):
        return None

    def _build(self, title=None, **suggested_paras):

        suggested_paras = jsonize(suggested_paras)

        verbosity = max(self.verbosity-1, 0)
        if 'verbosity' in self.model_kws:
            verbosity = self.model_kws.pop('verbosity')

        model = Model(
            model = {"layers": suggested_paras.pop('layers')},
             prefix=title,
             verbosity=verbosity,
            **suggested_paras,
            **self.model_kws
            )
        
        setattr(self, 'model_', model)

        return model

    def build_from_config(self, data, config_path, weight_file, **kwargs):

        model = Model.from_config_file(config_path=config_path)
        weight_file = os.path.join(model.w_path, weight_file)
        model.update_weights(weight_file=weight_file)

        test_true, test_pred = model.predict(data=data, return_true=True)

        train_true, train_pred = model.predict(data='training', return_true=True)

        setattr(self, 'model_', model)

        return (train_true, train_pred), (test_true, test_pred)

    def model_Dense(self, **kwargs):
        # Dense 
        self.param_space = self._make_param_space([Integer(16, 100, name='dense_units')])

        self.x0 = self._make_x0([32])

        _layers = self._make_layers({
            "Dense": kwargs.get('dense_units', 32)
            })

        return self._make_return(_layers, **kwargs)

    def model_DenseAct(self, **kwargs):
        # Dense, Activation
        self.param_space = self._make_param_space([
            Integer(16, 100, name='dense_units'),
            Categorical(['relu', 'sigmoid', 'elu', 'tanh'], name="activation")
            ])

        self.x0 = self._make_x0([32, 'relu'])

        _layers = self._make_layers({
            "Dense": kwargs.get('dense_units', 32),
            "Activation": {"activation": kwargs.get('activation', 'relu')},
            })

        return self._make_return(_layers, **kwargs)

    def model_DenseActDense(self, **kwargs):
        # Dense, Activation, Dense
        self.param_space = self._make_param_space([
            Integer(16, 100, name='dense_units'),
            Categorical(['relu', 'sigmoid', 'elu', 'tanh'], name="activation"),
            Integer(10, 60, name='dense_units2')
            ])

        self.x0 = self._make_x0([32, 'relu', 16])

        _layers = self._make_layers({
            "Dense": kwargs.get('dense_units', 32),
            "Activation": {"activation": kwargs.get('activation', 'relu')},
            "Dense": kwargs.get('dense_units2', 16)
            })

        return self._make_return(_layers, **kwargs)

    def model_DenseActDenseAct(self, **kwargs):
        # Dense, Activation, Dense
        self.param_space = self._make_param_space([
            Integer(16, 100, name='dense_units'),
            Categorical(['relu', 'sigmoid', 'elu', 'tanh'], name="activation"),
            Integer(10, 60, name='dense_units2'),
            Categorical(['relu', 'sigmoid', 'elu', 'tanh'], name="activation2")
            ])

        self.x0 = self._make_x0([32, 'relu', 16, 'relu'])

        _layers = self._make_layers({
            "Dense": kwargs.get('dense_units', 32),
            "Activation": {"activation": kwargs.get('activation', 'relu')},
            "Dense": kwargs.get('dense_units2', 16),
            "Activation": {"activation": kwargs.get('activation2', 'relu')}
            })

        return self._make_return(_layers, **kwargs)

    def model_Lstm(self, **kwargs):
        # lstm, dense
        self.param_space = self._make_param_space([Integer(16, 100, name='lstm_units')])

        self.x0 = self._make_x0([32])

        _layers = self._make_layers({
            "LSTM": kwargs.get('lstm_units', 32)
            })

        return self._make_return(_layers, **kwargs)

    def model_LstmAct(self, **kwargs):
        # lstm, Activation
        self.param_space = self._make_param_space([
            Integer(16, 100, name='lstm_units'),
            Categorical(['relu', 'sigmoid', 'elu', 'tanh'], name="activation")
            ])

        self.x0 = self._make_x0([32, 'relu'])

        _layers = self._make_layers({
            "LSTM": {"units": kwargs.get('lstm_units', 32),},
            "Activation": {"activation": kwargs.get('activation', 'relu')},
                  })

        return self._make_return(_layers, **kwargs)

    def model_LstmActLstm(self, **kwargs):
        # lstm, Activation, lstm
        self.param_space = self._make_param_space([
            Integer(16, 100, name='lstm_units'),
            Categorical(['relu', 'sigmoid', 'elu', 'tanh'], name="activation"),
            Integer(10, 60, name='lstm_units2')
            ])

        self.x0 = self._make_x0([32, 'relu', 16])

        _layers = self._make_layers({
            "LSTM": {"units": kwargs.get('lstm_units', 32), "return_sequences": True},
            "Activation": {"activation": kwargs.get('activation', 'relu')},
            "LSTM": kwargs.get('lstm_units2', 16)
                  })

        return self._make_return(_layers, **kwargs)

    def model_Cnn1D(self, **kwargs):

        self.param_space = self._make_param_space([
            Integer(16, 100, name='filters'),
            Integer(3, 5, name='kernel_size')
        ])

        self.x0 = [32, 5]

        _layers = self._make_layers({
            "Conv1D": {"filters": kwargs.get('filters', 32), 
                       "kernel_size": kwargs.get('kernel_size', 5)},
            "MaxPool1D": 2,
            })

        return self._make_return(_layers, **kwargs)

    def model_Cnn1DAct(self, **kwargs):

        self.param_space = self._make_param_space([
            Integer(16, 100, name='filters'),
            Integer(3, 5, name='kernel_size'),
            Categorical(['relu', 'sigmoid', 'elu', 'tanh'], name="activation")
        ])

        self.x0 = [32, 5, 'relu']

        _layers = self._make_layers({
            "Conv1D": {"filters": kwargs.get('filters', 32), 
                       "kernel_size": kwargs.get('kernel_size', 5)},
            "MaxPool1D": 2,
            "Activation": {"activation": kwargs.get('activation', 'relu')},
            })

        return self._make_return(_layers, **kwargs)

    def _make_param_space(self, space:list)->list:
        """makes a parameter space"""
        return space + self.static_space

    def _make_x0(self, x0)->list:
        """makes initial/starting parameters"""  
        return x0 + self.static_x0

    def _make_layers(self, layers:dict)->dict:

        layers.update(self._output_layer())

        return layers
    
    def _make_return(self, layers, **kwargs)->dict:
        
        return {'layers': layers,    
            'ts_args': {'lookback': kwargs.get('lookback', self.model_kws.get('lookback', 8))},
            'batch_size': kwargs.get('batch_size', self.model_kws.get('batch_size', 8)),
            'lr': kwargs.get('lr', self.model_kws.get('batch_size', 0.001)),
            }
    
    def _output_layer(self)->dict:
        return {
            "Flatten_out": {},
            "Dense_out": 1
            }




import os
import random
import warnings
from typing import Union

import numpy as np
from easy_mpl import imshow
import matplotlib.pyplot as plt
from sklearn.tree import plot_tree

from ai4water.backend import tf, keras

if tf is not None:
    import ai4water.keract_mod as keract
else:
    keract = None

from ai4water.utils.plotting_tools import Plots
from ai4water.utils.utils import maybe_three_outputs, get_nrows_ncols

from ..utils import choose_examples

try:
    from ai4water.utils.utils_from_see_rnn import rnn_histogram
except ModuleNotFoundError:
    rnn_histogram = None


try:
    from dtreeviz import trees
except ModuleNotFoundError:
    trees = None

try:
    import lightgbm
except ModuleNotFoundError:
    lightgbm = None

try:
    import xgboost
except ModuleNotFoundError:
    xgboost = None

RNN_INFO = {"LSTM": {'rnn_type': 'LSTM',
                     'gate_names': ['INPUT', 'FORGET', 'CELL', 'OUTPUT'],
                     'n_gates': 4,
                     'is_bidir': False,
                     'rnn_dim': 64,
                     'uses_bias': True,
                     'direction_names': [[]]
                     }
            }

CMAPS = [
    'jet_r', 'ocean_r', 'viridis_r', 'BrBG',
    'GnBu',
    #'crest_r',
    'Blues_r', 'bwr_r',
    #'flare',
    'YlGnBu'
]


TREE_BASED_MODELS = [
    "DecisionTreeRegressor",
    "ExtraTreeRegressor",
    "XGBRFRegressor",
    "XGBRegressor",
    "CatBoostRegressor",
    "LGBMRegressor",

    "DecisionTreeClassifier",
    "ExtraTreeClassifier",
    "XGBClassifier",
    "XGBRFClassifier",
    "CatBoostClassifier",
    "LGBMClassifier"

]


class Visualize(Plots):
    """Hepler class to peek inside the machine learning mdoel.

    If the machine learning model consists of layers of neural networks,
    then this class can be used to plot following 4 items

        - outputs of individual layers
        - gradients of outputs of individual layers
        - weights and biases of individual layers
        - gradients of weights of individual layers

    If the machine learning model consists of tree, then this
    class can be used to plot the learned tree of the model.

    methods
    -------
        - get_activations
        - activations
        - get_activation_gradients
        - activation_gradients
        - get_weights
        - weights
        - get_weight_gradients
        - weight_gradients
        - decision_tree
    """

    def __init__(
            self,
            model
    ):
        """
        Arguments:
            model : the learned machine learning model.
        """
        plt.rcParams.update(plt.rcParamsDefault)

        self.model = model
        self.verbosity = model.verbosity

        self.vis_path = os.path.join(model.path, "visualize")
        if not os.path.exists(self.vis_path):
            os.makedirs(self.vis_path)

        Plots.__init__(self,
                       path=self.vis_path,
                       config=model.config)

    def __call__(
            self,
            layer_name,
            data='training',
            x=None,
            y=None,
            examples_to_use=None,
            show: bool = False,
    ):

        if self.model.category == "DL":
            self.activations(layer_name, data, x, examples_to_use, show=show)
            self.activation_gradients(layer_name, x=x, y=y,
                                      examples_to_use=examples_to_use, show=show)
            self.weights(layer_name, show=show)
            self.weight_gradients(layer_name, data=data, x=x, y=y, show=show)
        else:
            self.decision_tree(show=show)
            self.decision_tree_leaves(data=data)

        return

    def get_activations(self,
                        layer_names: Union[list, str] = None,
                        x=None,
                        data: str = 'training',
                        batch_size=None,
                        ) -> dict:
        """gets the activations/outputs of any layer of the Keras Model.

        Arguments:
            layer_names : name of list of names of layers whose activations are
                to be returned.
            x : If provided, it will override `data`.
            data : data to use to get activations. Only relevent if `x` is not
                provided. By default training data is used. Possible values are
                `training`, `test` or `validation`.
        Returns:
            a dictionary whose keys are names of layers and values are weights
            of those layers as numpy arrays
        """
        if x is None:
            # if layer names are not specified, this will get get activations
            # of allparameters
            x, _ = getattr(self.model, f'{data}_data')()
            #x, _ = maybe_three_outputs(data, self.model.teacher_forcing)

        if self.model.api == "subclassing":
            dl_model = self.model
        else:
            dl_model = self.model._model

        if isinstance(x, list):
            num_examples = len(x[0])
        elif isinstance(x, np.ndarray):
            num_examples = len(x)
        else:
            raise ValueError

        if batch_size:
            # feed each batch and get activations per batch
            assert isinstance(layer_names, str)
            _activations = []
            for batch in range(num_examples // batch_size):
                batch_x = _get_batch_input(x, batch, batch_size)
                batch_activations = keract.get_activations(dl_model, batch_x, layer_names=layer_names,
                                                 auto_compile=True)
                assert len(batch_activations) == 1  # todo
                _activations.append(list(batch_activations.values())[0])
            activations = {layer_names: np.concatenate(_activations)}
        else:
            activations = keract.get_activations(dl_model, x, layer_names=layer_names,
                                             auto_compile=True)

        return activations

    def activations(
            self,
            layer_names=None,
            data: str = 'training',
            x=None,
            examples_to_use: Union[int, list, np.ndarray, range] = None,
            show: bool = False
    ):
        """Plots outputs of any layer of neural network.

        Arguments:
            data : The data to be used for calculating outputs of layers.
            x : if given, will override, 'data'.
            layer_names : name of layer whose output is to be plotted. If None,
                it will plot outputs of all layers
            examples_to_use : If integer, it will be the number of examples to use.
                If array like, it will be the indices of examples to use.
            show :
        """
        activations = self.get_activations(x=x, data=data)

        if layer_names is not None:
            if isinstance(layer_names, str):
                layer_names = [layer_names]
            else:
                layer_names = layer_names
        else:
            layer_names = list(activations.keys())

        assert isinstance(layer_names, list)

        if self.verbosity > 0:
            print("Plotting activations of layers")

        for lyr_name, activation in activations.items():

            if lyr_name in layer_names:
                # activation may be tuple e.g if input layer receives more than
                # 1 input
                if isinstance(activation, np.ndarray):

                    if activation.ndim == 2 and examples_to_use is None:
                        examples_to_use = len(activation)

                    self._plot_activations(activation, lyr_name, examples_to_use,
                                           show)

                elif isinstance(activation, tuple):
                    for act in activation:
                        self._plot_activations(act, lyr_name, show)
        return

    def _plot_activations(self, activation, lyr_name, examples_to_use=24, show=False):

        if examples_to_use is None:
            indices = range(len(activation))
        else:
            activation, indices = choose_examples(activation, examples_to_use)

        kwargs = {}

        if "LSTM" in lyr_name.upper() and np.ndim(activation) in (2, 3):

            if activation.ndim == 3:

                self.features_2d(activation,
                                 show=show,
                                 name=lyr_name + "_outputs",
                                 sup_title="Activations",
                                 n_rows=6,
                                 sup_xlabel="LSTM units",
                                 sup_ylabel="Lookback steps",
                                 title=indices,
                                 )
            else:
                self._imshow(activation, f"{lyr_name} Activations", fname=lyr_name,
                             show=show,
                             ylabel="Examples", xlabel="LSTM units",
                             cmap=random.choice(CMAPS))

        elif np.ndim(activation) == 2 and activation.shape[1] > 1:
            if "lstm" in lyr_name.lower():
                kwargs['xlabel'] = "LSTM units"
            self._imshow(activation, lyr_name + " Activations", show=show,
                         fname=lyr_name, **kwargs)

        elif np.ndim(activation) == 3:
            if "input" in lyr_name.lower():
                kwargs['xticklabels'] = self.model.input_features
            self._imshow_3d(activation, lyr_name, save=show, **kwargs, where='')
        elif np.ndim(activation) == 2:  # this is now 1d
            # shape= (?, 1)
            self.plot1d(activation, label=lyr_name + ' Outputs', show=show,
                        fname=lyr_name + '_outputs')
        else:
            print("ignoring activations for {} because it has shape {}, {}".format(lyr_name, activation.shape,
                                                                                   np.ndim(activation)))
        return

    def get_weights(self):
        """ returns all trainable weights as arrays in a dictionary"""
        weights = {}
        for weight in self.model.trainable_weights:
            if tf.executing_eagerly():
                weights[weight.name] = weight.numpy()
            else:
                weights[weight.name] = keras.backend.eval(weight)
        return weights

    def weights(
            self,
            layer_names: Union[str, list] = None,
            show: bool = False
    ):
        """Plots the weights of a specific layer or all layers.

        Arguments:
            layer_names : The layer whose weights are to be viewed.
            show :
        """

        weights = self.get_weights()

        if self.verbosity > 0:
            print("Plotting trainable weights of layers of the model.")

        if layer_names is None:
            layer_names = list(weights.keys())
        elif isinstance(layer_names, str):
            layer_names = [layer_names]
        else:
            layer_names = layer_names

        for lyr in layer_names:

            for _name, weight in weights.items():

                if lyr in _name:
                    title = _name
                    fname = _name + '_weights'

                    rnn_args = None
                    if "LSTM" in title.upper():
                        rnn_args = {'n_gates': 4,
                                    'gate_names_str': "(input, forget, cell, output)"}

                    if np.ndim(weight) == 2 and weight.shape[1] > 1:

                        self._imshow(weight, title, show=show, fname=fname,
                                     rnn_args=rnn_args)

                    elif len(weight) > 1 and np.ndim(weight) < 3:
                        self.plot1d(weight, title, show, fname, rnn_args=rnn_args)

                    elif "conv" in _name.lower() and np.ndim(weight) == 3:
                        _name = _name.replace("/", "_")
                        _name = _name.replace(":", "_")

                        self.features_2d(data=weight,
                                         save=show,
                                         name=_name,
                                         slices=64,
                                         slice_dim=2,
                                         tight=True,
                                         borderwidth=1)
                    else:
                        print("ignoring weight for {} because it has shape {}".format(_name, weight.shape))
        return

    def get_activation_gradients(
            self,
            layer_names: Union[str, list] = None,
            x=None,
            y=None,
            data: str = 'training'
    ) -> dict:
        """
        Finds gradients of outputs of a layer.

        either x,y or data is required
        Arguments:
            layer_names : The layer for which, the gradients of its outputs are to be
                calculated.
            x : input data. Will overwrite `data`
            y : corresponding label of x. Will overwrite `data`.
            data : one of `training`, `test` or `validation`
        """
        if isinstance(layer_names, str):
            layer_names = [layer_names]

        if x is None:
            data = getattr(self.model, f'{data}_data')()
            x, y = maybe_three_outputs(data)

        return keract.get_gradients_of_activations(self.model, x, y,
                                                   layer_names=layer_names)

    def activation_gradients(
            self,
            layer_names: Union[str, list],
            data='training',
            x=None,
            y=None,
            examples_to_use=None,
            plot_type="2D",
            show: bool = False
    ):
        """Plots the gradients o activations/outputs of layers

        Arguments:
            layer_names : the layer name for which the gradients of its outputs
                are to be plotted.
            data : the data to be used for calculating gradients
            x : alternative to data
            y : alternative to data
            examples_to_use : the examples from the data to use. If None, then all
                examples will be used, which is equal to the length of data.
            plot_type :
            show :
        """
        if plot_type == "2D":
            return self.activation_gradients_2D(layer_names, data, x, y,
                                                examples_to_use, show)

        return self.activation_gradients_1D(layer_names, data, x, y, examples_to_use,
                                            show)

    def activation_gradients_2D(self,
                                layer_names=None,
                                data='training',
                                x=None,
                                y=None,
                                examples_to_use=24,
                                show=True
                                ):
        """Plots activations of intermediate layers except input and output

        Arguments:
            layer_names :
            data :
            x :
            y :
            examples_to_use : if integer, it will be the number of examples to use.
                If array like, it will be index of examples to use.
            show :
        """

        gradients = self.get_activation_gradients(layer_names=layer_names,
                                                  data=data, x=x, y=y)

        return self._plot_act_grads(gradients, examples_to_use, show=show)

    def activation_gradients_1D(self,
                                layer_names,
                                data='training',
                                x=None,
                                y=None,
                                examples_to_use=None,
                                show=False):
        """Plots gradients of layer outputs as 1D

        Arguments:
            layer_names :
            examples_to_use :
            data :
            x :
            y :
            show :
        """
        gradients = self.get_activation_gradients(layer_names=layer_names,
                                                  data=data, x=x, y=y)

        for lyr_name, gradient in gradients.items():
            fname = lyr_name + "_output_grads"
            title = lyr_name + " Output Gradients"

            if np.ndim(gradient) == 3:
                for idx, example in enumerate(gradient):

                    _title = f"{title}_{idx}"
                    _fname = f"{fname}_{idx}"

                    if "LSTM" in lyr_name:
                        example = example.T

                    self.features_1d(example, name=_fname, title=_title, show=show,
                                     xlabel="Lookback steps", ylabel="Gradients")
        return

    def _plot_act_grads(self, gradients, examples_to_use=24, show=True):

        if self.verbosity > 0:
            print("Plotting gradients of activations of layersr")

        for lyr_name, gradient in gradients.items():

            if examples_to_use is None:
                indices = range(len(gradient))
            else:
                gradient, indices = choose_examples(gradient, examples_to_use)

            fname = lyr_name + "_output_grads"
            title = lyr_name + " Output Gradients"
            if "LSTM" in lyr_name.upper() and np.ndim(gradient) in (2, 3):

                if gradient.ndim == 2:
                    self._imshow(gradient, fname=fname, label=title, show=show,
                                 xlabel="LSTM units")
                else:

                    self.features_2d(gradient,
                                     name=fname,
                                     title=indices,
                                     show=show,
                                     n_rows=6,
                                     sup_title=title,
                                     sup_xlabel="LSTM units",
                                     sup_ylabel="Lookback steps")

            elif np.ndim(gradient) == 2:
                if gradient.shape[1] > 1:
                    # (?, ?)
                    self._imshow(gradient, title, show, fname)
                elif gradient.shape[1] == 1:
                    # (? , 1)
                    self.plot1d(np.squeeze(gradient), title, show, fname)

            elif np.ndim(gradient) == 3 and gradient.shape[1] == 1:
                if gradient.shape[2] == 1:
                    # (?, 1, 1)
                    self.plot1d(np.squeeze(gradient), title, show, fname)
                else:
                    # (?, 1, ?)
                    self._imshow(np.squeeze(gradient), title, show, fname)
            elif np.ndim(gradient) == 3:
                if gradient.shape[2] == 1:
                    # (?, ?, 1)
                    self._imshow(np.squeeze(gradient), title, show, fname)
                elif gradient.shape[2] > 1:
                    # (?, ?, ?)
                    self._imshow_3d(gradient, lyr_name, show)
            else:
                print("ignoring activation gradients for {} because it has shape {} {}".format(lyr_name, gradient.shape,
                                                                                               np.ndim(gradient)))

    def get_weight_gradients(
            self,
            data: str = 'training',
            x=None,
            y=None
    ) -> dict:
        """Returns the gradients of weights.

        Arguments:
            data : the data to use to calculate gradients of weights.
            x :
            y :

        Returns:
            dictionary whose keys are names of layers and values are gradients of
            weights as numpy arrays.
        """
        if x is None:
            data = getattr(self.model, f'{data}_data')()
            x, y = maybe_three_outputs(data)

        return keract.get_gradients_of_trainable_weights(self.model, x, y)

    def weight_gradients(
            self,
            layer_names: Union[str, list] = None,
            data='training',
            x=None,
            y=None,
            show: bool = False,
    ):
        """Plots gradient of all trainable weights

        Arguments:
            layer_names : the layer whose weeights are to be considered.
            data :  the data to use to calculate gradients of weights
            x : alternative to data
            y : alternative to data
            show : whether to show the plot or not.
        """
        gradients = self.get_weight_gradients(data=data, x=x, y=y)

        if layer_names is None:
            layers_to_plot = list(gradients.keys())
        elif isinstance(layer_names, str):
            layers_to_plot = [layer_names]
        else:
            layers_to_plot = layer_names

        if self.verbosity > 0:
            print("Plotting gradients of trainable weights")

        for lyr_to_plot in layers_to_plot:

            for lyr_name, gradient in gradients.items():

                # because lyr_name is most likely larger
                if lyr_to_plot in lyr_name:

                    title = lyr_name + "Weight Gradients"
                    fname = lyr_name + '_weight_grads'
                    rnn_args = None

                    if "LSTM" in title.upper():
                        rnn_args = {'n_gates': 4,
                                    'gate_names_str': "(input, forget, cell, output)"}

                        if np.ndim(gradient) == 3:
                            self.rnn_histogram(gradient, name=fname, title=title,
                                               show=show)

                    if np.ndim(gradient) == 2 and gradient.shape[1] > 1:
                        self._imshow(gradient, title, show=show, fname=fname,
                                     rnn_args=rnn_args)

                    elif len(gradient) and np.ndim(gradient) < 3:
                        self.plot1d(gradient, title, show=show, fname=fname,
                                    rnn_args=rnn_args)
                    else:
                        print(f"""ignoring weight gradients for {lyr_name} because it has
                         shape {gradient.shape} {np.ndim(gradient)}""")
        return

    def find_num_lstms(self, layer_names=None) -> list:
        """Finds names of lstm layers in model"""
        if layer_names is not None:
            if isinstance(layer_names, str):
                layer_names = [layer_names]
            assert isinstance(layer_names, list)

        lstm_names = []
        for lyr, config in self.config['model']['layers'].items():
            if "LSTM" in lyr.upper():
                config = config.get('config', config)
                prosp_name = config.get('name', lyr)

                if layer_names is not None:
                    if prosp_name in layer_names:
                        lstm_names.append(prosp_name)
                else:
                    lstm_names.append(prosp_name)

        return lstm_names

    def get_rnn_weights(self, weights: dict, layer_names=None) -> dict:
        """Finds RNN related weights.

         It combines kernel recurrent curnel and bias of each layer into a list."""
        lstm_weights = {}
        if self.config['model'] is not None and 'layers' in self.config['model']:
            if "LSTM" in self.config['model']['layers']:
                lstms = self.find_num_lstms(layer_names)
                for lstm in lstms:
                    lstm_w = []
                    for w in ["kernel", "recurrent_kernel", "bias"]:
                        w_name = lstm + "/lstm_cell/" + w
                        w_name1 = f"{lstm}/{w}"
                        for k, v in weights.items():
                            if any(_w in k for _w in [w_name, w_name1]):
                                lstm_w.append(v)

                    lstm_weights[lstm] = lstm_w

        return lstm_weights

    def rnn_weights_histograms(self, layer_name, show=False):

        weights = self.get_weights()
        rnn_weights = self.get_rnn_weights(weights, layer_name)

        for k, w in rnn_weights.items():
            self.rnn_histogram(w, name=k + "_weight_histogram", show=show)

        return

    def rnn_weight_grads_as_hist(self,
                                 layer_name=None,
                                 data='training',
                                 x=None,
                                 y=None,
                                 show=False
                                 ):

        gradients = self.get_weight_gradients(data=data, x=x, y=y)

        rnn_weights = self.get_rnn_weights(gradients)
        for k, w in rnn_weights.items():
            self.rnn_histogram(w, name=k + "_weight_grads_histogram", show=show)

        return

    def rnn_histogram(self, data, save=True, name='', show=False, **kwargs):

        if save:
            save = os.path.join(self.vis_path, name + "0D.png")
        else:
            save = None

        if rnn_histogram is None:
            warnings.warn("install see-rnn to plot rnn_histogram plot", UserWarning)
        else:
            rnn_histogram(data, RNN_INFO["LSTM"], bins=400, savepath=save,
                          show=show, **kwargs)

        return

    def decision_tree(self, show=False, **kwargs):
        """Plots the decision tree"""

        fname = os.path.join(self.path, "decision_tree")
        if self.model.category == "ML":
            model_name = list(self.model.config['model'].keys())[0]

            if model_name in TREE_BASED_MODELS:

                _fig, axis = plt.subplots(figsize=kwargs.get('figsize', (10, 10)))

                if model_name.startswith("XGB"):
                    # https://xgboost.readthedocs.io/en/latest/python/python_api.html#xgboost.plot_tree
                    xgboost.plot_tree(self.model._model, ax=axis, **kwargs)

                elif model_name.startswith("Cat"):

                    gv_object = self.model._model.plot_tree(0, **kwargs)
                    if show:
                        gv_object.view()
                    gv_object.save(filename="decision_tree", directory=self.path)

                elif model_name.startswith("LGBM"):
                    lightgbm.plot_tree(self.model._model, ax=axis, **kwargs)

                else:  # sklearn types
                    plot_tree(self.model._model,
                              feature_names=self.model.input_features,
                              ax=axis, **kwargs)

                plt.savefig(fname, dpi=500)
                if show:
                    plt.show()
            else:
                print(f"decision tree can not be plotted for {model_name}")
        else:
            print(f"decision tree can not be plotted for {self.model.category} models")
        return

    def decision_tree_leaves(self, save=True, data='training'):
        """Plots dtreeviz related plots if dtreeviz is installed"""

        model = list(self.config['model'].keys())[0]
        if model in ["DecisionTreeRegressor", "DecisionTreeClassifier"]:

            if trees is None:
                print("dtreeviz related plots can not be plotted")
            else:
                x, y = getattr(self.model, f'{data}_data')()

                if np.ndim(y) > 2:
                    y = np.squeeze(y, axis=2)

                trees.viz_leaf_samples(self.model._model, x, self.in_cols)
                self.save_or_show(save, fname="viz_leaf_samples", where="plots")

                trees.ctreeviz_leaf_samples(self.model._model, x, y,
                                            self.in_cols)
                self.save_or_show(save, fname="ctreeviz_leaf_samples", where="plots")
        return

    def features_2d(self, data, name, save=True, slices=24, slice_dim=0, **kwargs):
        """Calls the features_2d from see-rnn"""
        st=0
        if 'title' in kwargs:
            title = kwargs.pop('title')
        else:
            title = None

        for en in np.arange(slices, data.shape[slice_dim] + slices, slices):

            if save:
                fname = name + f"_{st}_{en}"
                save = os.path.join(self.path, fname+".png")
            else:
                save = None

            if isinstance(title, np.ndarray):
                _title = title[st:en]
            else:
                _title = title

            if slice_dim == 0:
                features_2D(data[st:en, :], savepath=save, title=_title, **kwargs)
            else:
                # assuming it will always be the last dim if not first
                features_2D(data[..., st:en], savepath=save, title=_title, **kwargs)
            st = en
        return

    def features_1d(self, data, save=True, name='', **kwargs):

        if save:
            save = os.path.join(self.path, name + ".png")
        else:
            save=None

        if features_1D is None:
            warnings.warn("install see-rnn to plot features-1D plot", UserWarning)
        else:
            features_1D(data, savepath=save, **kwargs)

        return


def features_2D(data,
                n_rows=None,
                cmap=None,
                sup_xlabel=None,
                sup_ylabel=None,
                sup_title=None,
                title=None,
                show=False,
                savepath=None):
    """
    title: title for individual axis
    sup_title: title for whole plot
    """

    n_subplots = len(data) if data.ndim == 3 else 1
    nrows, ncols = get_nrows_ncols(n_rows, n_subplots)

    cmap = cmap or random.choice(CMAPS)

    fig, axis = plt.subplots(nrows=nrows, ncols=ncols,
                             dpi=100, figsize=(10, 10),
                             sharex='all', sharey='all')

    num_subplots = len(axis.ravel()) if isinstance(axis, np.ndarray) else 1

    if isinstance(title, str):
        title = [title for _ in range(num_subplots)]
    elif isinstance(title, list):
        assert len(title) == num_subplots
    elif isinstance(title, np.ndarray):
        assert len(title) == num_subplots
    elif title:
        title = np.arange(num_subplots)

    if isinstance(axis, plt.Axes):
        axis = np.array([axis])

    vmin = data.min()
    vmax = data.max()

    for idx, ax in enumerate(axis.flat):
        ax, im = imshow(data[idx],
                          ax=ax,
                          cmap=cmap, vmin=vmin, vmax=vmax,
                          title=title[idx],
                          show=False)

    if sup_xlabel:
        fig.text(0.5, 0.04, sup_xlabel, ha='center', fontsize=20)

    if sup_ylabel:
        fig.text(0.04, 0.5, sup_ylabel, va='center', rotation='vertical',
                 fontsize=20)

    fig.subplots_adjust(hspace=0.2)

    cbar_ax = fig.add_axes([0.92, 0.15, 0.05, 0.7])
    fig.colorbar(im, cax=cbar_ax)
    if sup_title:
        plt.suptitle(sup_title)

    # fig.tight_layout()

    if savepath:
        plt.savefig(savepath, bbox_inches="tight")

    if show:
        plt.show()
    return


def features_1D(data, xlabel=None, ylabel=None, savepath=None, show=None, 
        title=None):

    assert data.ndim == 2

    _, axis = plt.subplots()

    for i in data:
        axis.plot(i)

    if xlabel:
        axis.set_xlabel(xlabel)

    if ylabel:
        axis.set_ylabel(ylabel)

    if title:
        axis.set_title(title)

    if savepath:
        plt.savefig(savepath, bbox_inches="tight")

    if show:
        plt.show()
    return


def _get_batch_input(inp, batch_idx, batch_size):
    if isinstance(inp, list):
        batch_inp = []
        for x in inp:
            st = batch_idx*batch_size
            batch_inp.append(x[st: st+batch_size])
    else:
        st = batch_idx * batch_size
        batch_inp = inp[st: st+batch_size]
    return batch_inp



__all__ = ["HARHNModel", "IMVModel"]

import os
from typing import Any

import numpy as np
import matplotlib as mpl
from easy_mpl import imshow
import matplotlib.pyplot as plt

from .main import Model
from .utils.utils import dateandtime_now, plot_activations_along_inputs
from .backend import torch

if torch is not None:
    from .models.torch import IMVTensorLSTM
    from .models.torch import HARHN
else:
    HARHN, IMVTensorLSTM = None, None


class HARHNModel(Model):

    def __init__(self, use_gpu=True, teacher_forcing=True, **kwargs):

        dev = torch.device("cpu")
        if use_gpu and torch.cuda.is_available():
            dev = torch.device("cuda")

        self.dev = dev

        super(HARHNModel, self).__init__(teacher_forcing=teacher_forcing, **kwargs)

        # should be set after initiating upper classes so that torch_learner attribute is set
        self.torch_learner.use_cuda = use_gpu

    def initialize_layers(self, layers_config: dict, inputs=None):

        self.pt_model = HARHN(layers_config['n_conv_lyrs'],
                              self.lookback,
                              self.num_ins,
                              self.num_outs,
                              n_units_enc=layers_config['enc_units'],
                              n_units_dec=layers_config['dec_units'],
                              use_predicted_output=self.teacher_forcing,  # self.config['use_predicted_output']
                              ).to(self.dev)

        return

    def forward(self, *inputs: Any, **kwargs: Any):
        y_pred, _ = self.pt_model(inputs[0], inputs[1][:, -1], **kwargs)
        return y_pred


class IMVModel(HARHNModel):

    def __init__(self, *args, teacher_forcing=False, **kwargs):
        super(IMVModel, self).__init__(*args, teacher_forcing=teacher_forcing, **kwargs)

    def initialize_layers(self, layers_config:dict, inputs=None):

        self.pt_model = IMVTensorLSTM(self.num_ins, self.num_outs,
                                      layers_config['hidden_units'],
                                      device=self.dev).to(self.dev)
        self.alphas, self.betas = [], []

        return

    def forward(self, *inputs: Any, **kwargs: Any):
        y_pred, alphas, betas = self.pt_model(*inputs, **kwargs)
        self.alphas.append(alphas)
        self.betas.append(betas)
        return y_pred

    def interpret(self,
                  data='training',
                  x=None,
                  annotate=True,
                  vmin=None,
                  vmax=None,
                  **bar_kws,
                  ):

        mpl.rcParams.update(mpl.rcParamsDefault)

        self.alphas, self.betas = [], []
        true, predicted = self.predict(data=data, process_results=False, return_true=True)

        name = f'data_on_{dateandtime_now()}' if x is not None else data

        betas = [array.detach().cpu().numpy() for array in self.betas]
        betas = np.concatenate(betas)  # (examples, ins, 1)
        betas = betas.mean(axis=0)  # (ins, 1)
        betas = betas[..., 0]  # (ins, )

        alphas = [array.detach().cpu().numpy() for array in self.alphas]
        alphas = np.concatenate(alphas)  # (examples, lookback, ins, 1)

        x, _ = getattr(self, f'{data}_data')()

        path = os.path.join(self.path, "interpret")
        if not os.path.exists(path):
            os.makedirs(path)

        plot_activations_along_inputs(data=x[:, -1, :],  # todo, is -1 correct?
                                      activations=alphas.reshape(-1, self.lookback, self.num_ins),
                                      observations=true,
                                      predictions=predicted,
                                      in_cols=self.input_features,
                                      out_cols=self.output_features,
                                      lookback=self.lookback,
                                      name=name,
                                      path=path,
                                      vmin=vmin,
                                      vmax=vmax
                                      )

        alphas = alphas.mean(axis=0)   # (lookback, ins, 1)
        alphas = alphas[..., 0]  # (lookback, ins)
        alphas = alphas.transpose(1, 0)  # (ins, lookback)

        all_cols = self.input_features
        plt.close('all')
        fig, ax = plt.subplots()
        fig.set_figwidth(16)
        fig.set_figheight(16)
        xticklabels=["t-"+str(i) for i in np.arange(self.lookback, 0, -1)]
        imshow(alphas,
               ax=ax,
               xticklabels=xticklabels,
               yticklabels=list(all_cols),
               title="Importance of features and timesteps",
               annotate=annotate,
               show=False)

        plt.savefig(os.path.join(path, f'acts_{name}'), dpi=400, bbox_inches='tight')

        plt.close('all')
        plt.bar(range(self.num_ins), betas, **bar_kws)
        plt.xticks(ticks=range(len(all_cols)), labels=list(all_cols), rotation=90, fontsize=12)
        plt.savefig(os.path.join(path, f'feature_importance_{name}'), dpi=400, bbox_inches='tight')
        return



import os

import numpy as np
import pandas as pd


class ExplainerMixin(object):

    def __init__(
            self,
            path,
            data,
            features
    ):
        if not os.path.exists(path):
            os.makedirs(path)

        self.path = path
        self.data = data
        self.features = features

    @property
    def data_is_2d(self):
        if isinstance(self.data, np.ndarray) and self.data.ndim == 2:
            return True
        elif isinstance(self.data, pd.DataFrame):
            return True
        else:
            return False

    @property
    def data_is_3d(self):
        if isinstance(self.data, np.ndarray) and self.data.ndim == 3:
            return True
        return False

    @property
    def single_source(self):
        if isinstance(self.data, list) and len(self.data) > 1:
            return False
        else:
            return True

    @property
    def features(self):
        return self._features

    @features.setter
    def features(self, features):
        if self.data_is_2d:
            if type(self.data) == pd.DataFrame:

                features = self.data.columns.to_list()
            elif features is None:
                features = [f"Feature {i}" for i in range(self.data.shape[-1])]
            else:
                assert isinstance(features, list) and len(features) == self.data.shape[-1], f"""
                    features must be given as list of length {self.data.shape[-1]} 
                    but are of len {len(features)}
                    """

                features = features
        elif not self.single_source and features is None:
            features = []
            for data in self.data:
                if isinstance(data, pd.DataFrame):
                    _features = data.columns.to_list()
                else:
                    _features = [f"Feature {i}" for i in range(data.shape[-1])]

                features.append(_features)

        elif self.data_is_3d and features is None:
            features = [f"Feature {i}" for i in range(self.data.shape[-1])]

        self._features = features

    @property
    def unrolled_features(self):
        # returns the possible names of features if unrolled over time
        if not self.data_is_2d and self.single_source:
            features = self.features
            if features is None:
                features = [f"Feature_{i}" for i in range(self.data.shape[-1])]

            lookback = self.data.shape[1]
            features = [[f"{f}_{i}" for f in features] for i in range(lookback)]
            features = [item for sublist in features for item in sublist]
        else:
            features = None
        return features



import numpy as np
import pandas as pd

from .utils import Utils
from .global_variables import LAMBDA

# TODO, classify methods which require wind_speed, or which require solar_rad.

class ETBase(Utils):
    """
    This is the base class for evapotranspiration calculation. It calculates
    etp according to Jensen and Haise_ method. Any new ETP calculation must
    inherit from it and must implement
    the ``__call__`` method.

    .. _Haise:
        https://doi.org/10.1061/JRCEA4.0000287
    """
    def __init__(self,
                 input_df: pd.DataFrame,
                 units:dict,
                 constants:dict,
                 **kwargs
                 ):
        """
        Parameters
        ---------
            input_df :
            units :
            constants :
            kwargs :
        """
        self.name = self.__class__.__name__

        super(ETBase, self).__init__(input_df.copy(),
                                     units.copy(),
                                     constants.copy(),
                                     **kwargs)

    def requirements(self, **kwargs):

        if 'constants' in kwargs:
            constants = kwargs['constants']
        else:
            constants = ['lat_dec_deg', 'altitude', 'ct', 'tx']

        if 'ts' in kwargs:
            ts = kwargs['ts']
        else:
            ts = ['temp']

        for cons in constants:
            if cons not in self.cons:
                if cons in self.default_cons:
                    val = self.default_cons[cons]['def_val']
                    desc = self.default_cons[cons]['desc']
                    if val is not None:
                        print("Warning: default value {} of parameter {} which is {} is being used".format(val,
                                                                                                           cons,
                                                                                                           desc))
                        self.cons[cons] = val
                else:
                    raise ValueError("Value of constant {} must be provided to calculate ETP using {}"
                                     .format(cons, self.name))
        for _ts in ts:
            if _ts not in self.input.columns:
                raise ValueError("Timeseries {} is required for calculation of ETP using {}"
                                 .format(_ts, self.name))

    def __call__(self, *args,
                 transform: bool=False,
                 **kwargs):
        """
        as given (eq 9) in Xu and Singh, 2000 and implemented here_

        uses:  a_s, b_s, ct=0.025, tx=-3
        Arguments:
            transform : whether to transform the calculated etp to frequecies
                other than at which it is calculated.

        .. _here:
            https://github.com/DanluGuo/Evapotranspiration/blob/8efa0a2268a3c9fedac56594b28ac4b5197ea3fe/R/Evapotranspiration.R#L2734
        """
        self.requirements(constants=['lat_dec_deg', 'altitude', 'ct', 'tx'],
                          ts=['temp'])
        rs = self.rs()
        tmp1 = np.multiply(np.multiply(self.cons['ct'], np.add(self.input['temp'], self.cons['tx'])), rs)
        et = np.divide(tmp1, LAMBDA)

        self.post_process(et, transform=transform)
        return et

    def post_process(self, et, transform=False):
        if isinstance(et, np.ndarray):
            et = pd.Series(et, index=self.input.index)
        self.output['et_' + self.name + '_' + self.freq_str] = et
        if transform:
            self.transform_etp(self.name)

    def summary(self):

        methods_evaluated = []
        for m in self.output.keys():
            if 'Hourly' in m:
                methods_evaluated.append(m)

        for m in methods_evaluated:
            ts = self.output[m]
            yrs = np.unique(ts.index.year)
            print('For {} \n'.format(m.split('_')[1], end=','))
            for yr in yrs:
                st, en = str(yr) + '0101', str(yr) + '1231'
                yr_ts = ts[st:en]
                yr_sum = yr_ts.sum().values[0]
                yr_mean = yr_ts.mean().values[0]
                print('for year {}:, sum: {:<10.1f} mean: {:<10.1f}'.format(yr, yr_sum, yr_mean))


class Abtew(ETBase):
    """
    daily etp using equation 3 in Abtew_ 1996. `k` is a dimentionless coefficient.

     uses: , k=0.52, a_s=0.23, b_s=0.5
    :param `k` coefficient, default value taken from [1]
    :param `a_s` fraction of extraterrestrial radiation reaching earth on sunless days
    :param `b_s` difference between fracion of extraterrestrial radiation reaching full-sun days
             and that on sunless days.

    .. _Abtew:
        https://doi.org/10.1111/j.1752-1688.1996.tb04044.x

     """
    def __call__(self, *args, **kwargs):

        self.requirements(constants=['lat_dec_deg', 'altitude', 'abtew_k'])

        rs = self.rs()
        et = np.multiply(self.cons['abtew_k'], np.divide(rs, LAMBDA))

        self.post_process(et, kwargs.get('transform', False))
        return et


class Albrecht(ETBase):
    """
     Developed in Germany by Albrecht, 1950. Djaman et al., 2016 Wrote the formula as

     eto = (0.1005 + 0.297 * u2) * (es - ea)

    """
    def __call__(self, *args, **kwargs):

        # Mean saturation vapour pressure
        if 'es' not in self.input:
            if self.freq_str == 'Daily':
                es = self.mean_sat_vp_fao56()
            elif self.freq_str == 'Hourly':
                es = self.sat_vp_fao56(self.input['temp'].values)
            elif self.freq_str == 'sub_hourly':   # TODO should sub-hourly be same as hourly?
                es = self.sat_vp_fao56(self.input['temp'].values)
            else:
                raise NotImplementedError
        else:
            es = self.input['es']

        # actual vapour pressure
        ea = self.avp_from_rel_hum()

        u2 = self._wind_2m()
        eto = (0.1005 + 0.297 * u2) * (es - ea)

        self.post_process(eto, kwargs.get('transform', False))
        return eto


class BlaneyCriddle(ETBase):
    """
    using formulation of Blaney-Criddle for daily reference crop ETP using monthly
    mean tmin and tmax. Inaccurate under extreme climates. underestimates in
    windy, dry and sunny conditions and overestimates under calm, humid and
    clouded conditions.

    Doorenbos, J., & Pruitt, W. O. (1977). Crop water requirements, FAO
    Irrigation and Drainage. Paper 24, 2a ed., Roma, Italy.
    """
    def __call__(self, *args, **kwargs):
        # TODO include modified BlaneyCriddle as introduced  in [3]

        self.requirements(constants=['e0', 'e1', 'e2', 'e3', 'e4'])  # check that all constants are present

        N = self.daylight_fao56()  # mean daily percentage of annual daytime hours
        u2 = self._wind_2m()
        rh_min = self.input['rh_min']
        n = self.input['sunshine_hrs']
        ta = self.input['temp'].values
        # undefined working variable (Allena and Pruitt, 1986; Shuttleworth, 1992) (S9.8)
        a1 = self.cons['e0'] + self.cons['e1'] * rh_min + self.cons['e2'] * n / N
        a2 = self.cons['e3'] * u2
        a3 = self.cons['e4'] * rh_min * n / N + self.cons['e5'] * rh_min * u2
        bvar = a1 + a2 + a3
        # calculate yearly sum of daylight hours and assign that value to each point in array `N`
        n_annual = assign_yearly(N, self.input.index)

        # percentage of actual daytime hours for the day comparing to the annual sum of maximum sunshine hours
        p_y = 100 * n / n_annual['N'].values

        # reference crop evapotranspiration
        et = (0.0043 * rh_min - n / N - 1.41) + bvar * p_y * (0.46 * ta + 8.13)
        self.post_process(et, kwargs.get('transform', False))
        return et


class BrutsaertStrickler(ETBase):
    """
    using formulation given by BrutsaertStrickler

    :param `alpha_pt` Priestley-Taylor coefficient = 1.26 for Priestley-Taylor_
        model (Priestley and Taylor, 1972)
    :param `a_s` fraction of extraterrestrial radiation reaching earth on sunless days
    :param `b_s` difference between fracion of extraterrestrial radiation reaching full-sun days
             and that on sunless days.
    :param `albedo`  Any numeric value between 0 and 1 (dimensionless), albedo of the evaporative surface
            representing the portion of the incident radiation that is reflected back at the surface.
            Default is 0.23 for surface covered with short reference crop.
    :return: et

    .. _Priestley-Taylor:
        https://doi.org/10.1029/WR015i002p00443
    """
    def __call__(self, *args, **kwargs):
        self.requirements(constants=['alphaPT'])  # check that all constants are present

        delta = self.slope_sat_vp(self.input['temp'].values)
        gamma = self.psy_const()
        vabar = self.avp_from_rel_hum()  # Vapour pressure, *ea*
        vas = self.mean_sat_vp_fao56()
        u2 = self._wind_2m()
        f_u2 = np.add(2.626, np.multiply(1.381, u2))
        r_ng = self.net_rad(vabar)
        alpha_pt = self.cons['alphaPT']

        et = np.subtract(np.multiply(np.multiply((2*alpha_pt-1),
                                                 np.divide(delta, np.add(delta, gamma))),
                                     np.divide(r_ng, LAMBDA)),
                         np.multiply(np.multiply(np.divide(gamma, np.add(delta, gamma)), f_u2),
                                     np.subtract(vas, vabar)))
        self.post_process(et, kwargs.get('transform', False))
        return et


class Camargo(ETBase):
    """
    Originally presented by Camargo, 1971. Following formula is presented in
    Fernandes et al., 2012 quoting Sedyiama et al., 1997.

         eto = f * Tmean * ra * nd

    Gurski et al., 2018 has not written nd in formula. He expressed formula to
    convert extra-terresterial radiation into equivalent mm/day as
        ra[mm/day] = ra[MegaJoulePerMeterSquare PerDay] / 2.45
        where 2.45 is constant.

     eto: reference etp in mm/day.
     f: an empircal factor taken as 0.01
     ra: extraterrestrial radiation expressed as mm/day
     nd: length of time interval
    """
    def __call__(self, *args, **kwargs):
        self.requirements(constants=['f_camargo'])  # check that all constants are present

        ra = self._et_rad()
        if self.freq_str == 'Daily':
            ra = ra/2.45
        else:
            raise NotImplementedError
        et = self.cons['f_camargo'] * self.input['temp'] * ra
        self.post_process(et, kwargs.get('transform', False))
        return et


class Caprio(ETBase):
    """
    Developed by Caprio (1974). Pandey et al 2016 wrote the equation as
        eto = (0.01092708*t + 0.0060706) * rs
    """
    def __call__(self, *args, **kwargs):

        rs = self.rs()
        eto = (0.01092708 * self.input['temp'] + 0.0060706) * rs

        self.post_process(eto, kwargs.get('transform', False))
        return eto


class ChapmanAustralia(ETBase):
    """using formulation of Chapman_, 2001,

    uses: a_s=0.23, b_s=0.5, ap=2.4, alphaA=0.14, albedo=0.23

    .. _Chapman:
        https://116.90.59.164/MODSIM03/Volume_01/A03/04_Chapman.pdf
    """
    def __call__(self, *args, **kwargs):
        self.requirements(constants=['lat_dec_deg', 'altitude', 'alphaA', 'pan_ap', 'albedo'],
                          ts=['temp'])

        lat = self.cons['lat_dec_deg']
        a_p = 0.17 + 0.011 * abs(lat)
        b_p = np.power(10, (0.66 - 0.211 * abs(lat)))  # constants (S13.3)

        epan = self.evap_pan()

        et = np.add(np.multiply(a_p, epan), b_p)

        self.post_process(et, kwargs.get('transform', False))
        return et


class Copais(ETBase):
    """
    Developed for central Greece by Alexandris et al 2006 and used in Alexandris et al 2008.
    """
    def __call__(self, *args, **kwargs):
        et = None
        self.post_process(et, kwargs.get('transform', False))
        return et


class Dalton(ETBase):
    """
    using Dalton formulation as mentioned here_ in mm/dday

    uses:
      es: mean saturation vapour pressure
      ea: actual vapour pressure
      u2: wind speed


    .. _here:
        https://water-for-africa.org/en/dalton.html
    """
    def __call__(self, *args, **kwargs):

        u2 = self._wind_2m()
        fau = 0.13 + 0.14 * u2

        # Mean saturation vapour pressure
        if 'es' not in self.input:
            if self.freq_str == 'Daily':
                es = self.mean_sat_vp_fao56()
            elif self.freq_str == 'Hourly':
                es = self.sat_vp_fao56(self.input['temp'].values)
            elif self.freq_str == 'sub_hourly':   # TODO should sub-hourly be same as hourly?
                es = self.sat_vp_fao56(self.input['temp'].values)
            else:
                raise NotImplementedError
        else:
            es = self.input['es']

        # actual vapour pressure
        ea = self.avp_from_rel_hum()
        if 'vp_def' not in self.input:
            vp_d = es - ea   # vapor pressure deficit
        else:
            vp_d = self.input['vp_def']

        etp = fau * vp_d
        self.post_process(etp, kwargs.get('transform', False))
        return etp


class DeBruinKeijman(ETBase):
    """
     Calculates daily Pot ETP, developed by deBruin and Jeijman 1979 and used in Rosenberry et al 2004.
    """


class DoorenbosPruitt(ETBase):
    """
    Developed by Doorenbos and Pruitt (1777), Poyen et al wrote following equation
      et = a(delta/(delta+gamma) * rs) + b
      b = -0.3
      a = 1.066 - 0.13 x10^{-2} * rh + 0.045*ud - 0.2x10^{-3}*rh * ud - 0.315x10^{-4}*rh**2 - 0.11x10{-2}*ud**2
      used in Xu HP 2000.
    """


class GrangerGray(ETBase):
    """
    using formulation of Granger & Gray 1989 which is for non-saturated lands and modified form of penman 1948.

     uses: , wind_f`='pen48', a_s=0.23, b_s=0.5, albedo=0.23
    :param `wind_f` str, if 'pen48 is used then formulation of [1] is used otherwise formulation of [3] requires
             wind_f to be 2.626.
    :param `a_s fraction of extraterrestrial radiation reaching earth on sunless days
    :param `b_s` difference between fracion of extraterrestrial radiation reaching full-sun days
             and that on sunless days.
    :param `albedo`  Any numeric value between 0 and 1 (dimensionless), albedo of the evaporative surface
            representing the portion of the incident radiation that is reflected back at the surface.
            Default is 0.23 for surface covered with short reference crop.
    :return:

    https://doi.org/10.1016/0022-1694(89)90249-7

    """
    def __call__(self, *args, **kwargs):

        self.requirements(constants=['wind_f'])  # check that all constants are present

        if self.cons['wind_f'] not in ['pen48', 'pen56']:
            raise ValueError('value of given wind_f is not allowed.')

        if self.cons['wind_f'] == 'pen48':
            _a = 2.626
            _b = 0.09
        else:
            _a = 1.313
            _b = 0.06

        # rs = self.rs()
        delta = self.slope_sat_vp(self.input['temp'].values)
        gamma = self.psy_const()

        vabar = self.avp_from_rel_hum()  # Vapour pressure
        r_n = self.net_rad(vabar)   # net radiation
        vas = self.mean_sat_vp_fao56()

        u2 = self._wind_2m()
        fau = _a + 1.381 * u2
        ea = np.multiply(fau, np.subtract(vas, vabar))

        # dimensionless relative drying power  eq 7 in Granger, 1998
        dry_pow = np.divide(ea, np.add(ea, np.divide(np.subtract(r_n, self.soil_heat_flux()), LAMBDA)))
        # eq 6 in Granger, 1998
        g_g = 1 / (0.793 + 0.20 * np.exp(4.902 * dry_pow)) + 0.006 * dry_pow

        tmp1 = np.divide(np.multiply(delta, g_g), np.add(np.multiply(delta, g_g), gamma))
        tmp2 = np.divide(np.subtract(r_n, self.soil_heat_flux()), LAMBDA)
        tmp3 = np.multiply(np.divide(np.multiply(gamma, g_g), np.add(np.multiply(delta, g_g), gamma)), ea)
        et = np.add(np.multiply(tmp1, tmp2), tmp3)

        self.post_process(et, kwargs.get('transform', False))
        return et


class Hamon(ETBase):
    """calculates evapotranspiration in mm using Hamon 1963 method as given in Lu et al 2005. It uses daily mean
     temperature which can also be calculated
    from daily max and min temperatures. It also requires `daylight_hrs` which is hours of day light, which if not
    provided as input, will be calculated from latitutde. This means if `daylight_hrs` timeseries is not provided as
    input, then argument `lat` must be provided.

    pet = cts * n * n * vdsat
    vdsat = (216.7 * vpsat) / (tavc + 273.3)
    vpsat = 6.108 * exp((17.26939 * tavc)/(tavc + 237.3))

    :uses cts: float, or array of 12 values for each month of year or a time series of equal length as input data.
                 if it is float, then that value will be considered for whole year. Default value of 0.0055 was used
                 by Hamon 1961, although he later used different value but I am using same value as it is used by
                 WDMUtil. It should be also noted that 0.0055 is to be used when pet is in inches. So I am dividing
                 the whole pet by 24.5 in order to convert from inches to mm while still using 0.0055.

    References
    ----------
    Hamon, W. R. (1963). Computation of direct runoff amounts from storm rainfall. International Association of
    Scientific Hydrology Publication, 63, 52-62.

    Lu et al. (2005).  A comparison of six potential evaportranspiration methods for regional use in the
        southeastern United States.  Journal of the American Water Resources Association, 41, 621-633.
     """

    def __call__(self, *args, **kwargs):
        self.requirements(constants=['lat_dec_deg', 'altitude', 'albedo', 'cts'],
                          ts=['temp'])
        # allow cts to be provided as input while calling method, e.g we may want to use array
        if 'cts' in kwargs:
            cts = kwargs['cts']
        else:
            cts = self.cons['cts']

        if 'sunshine_hrs' not in self.input.columns:
            if 'daylight_hrs' not in self.input.columns:
                daylight_hrs = self.daylight_fao56()
            else:
                daylight_hrs = self.input['daylight_hrus']
            sunshine_hrs = daylight_hrs
            print('Warning, sunshine hours are consiered equal to daylight hours')
        else:
            sunshine_hrs = self.input['sunshine_hrs']

        sunshine_hrs = np.divide(sunshine_hrs, 12.0)

        # preference should be given to tmin and tmax if provided and if tmin, tmax is not provided then use temp which
        # is mean temperature. This is because in original equations, vd_sat is calculated as average of max vapour
        # pressure and minimum vapour pressue.
        if 'tmax' not in self.input.columns:
            if 'temp' not in self.input.columns:
                raise ValueError('Either tmax and tmin or mean temperature should be provided as input')
            else:
                vd_sat = self.sat_vp_fao56(self.input['temp'])
        else:
            vd_sat = self.mean_sat_vp_fao56()

        # in some literature, the equation is divided by 100 by then the cts value is 0.55 instead of 0.0055
        et = cts * 25.4 * np.power(sunshine_hrs, 2) * (216.7 * vd_sat * 10 / (np.add(self.input['temp'], 273.3)))

        self.post_process(et, kwargs.get('transform', False))
        return et


class HargreavesSamani(ETBase):
    """
    estimates daily ETo using Hargreaves method Hargreaves and Samani_, 1985.
    :uses
      temp
      tmin
      tmax
    :param
    method: str, if `1985`, then the method of 1985 (Hargreaves and Samani, 1985) is followed as calculated by and
    mentioned by Hargreaves and Allen, 2003.
    if `2003`, then as formula is used as mentioned in [1]
    Note: Current test passes for 1985 method.
    There is a variation of Hargreaves introduced by Trajkovic 2007 as mentioned in Alexandris 2008.

    .. _Samani:
        https://rdrr.io/cran/Evapotranspiration/man/ET.HargreavesSamani.html
    """

    def __call__(self, method='1985', **kwargs):
        self.requirements(constants=['lat_dec_deg', 'altitude', 'albedo'],
                          ts=['temp'])

        if method == '2003':
            tmp1 = np.multiply(0.0023, np.add(self.input['temp'], 17.8))
            tmp2 = np.power(np.subtract(self.input['tmax'].values, self.input['tmin'].values), 0.5)
            tmp3 = np.multiply(0.408, self._et_rad())
            et = np.multiply(np.multiply(tmp1, tmp2), tmp3)

        else:
            ra_my = self._et_rad()
            tmin = self.input['tmin'].values
            tmax = self.input['tmax'].values
            ta = self.input['temp'].values
            # empirical coefficient by Hargreaves and Samani (1985) (S9.13)
            c_hs = 0.00185 * np.power((np.subtract(tmax, tmin)), 2) - 0.0433 * (np.subtract(tmax, tmin)) + 0.4023
            et = 0.0135 * c_hs * ra_my / LAMBDA * np.power((np.subtract(tmax, tmin)), 0.5) * (np.add(ta, 17.8))

        self.post_process(et, kwargs.get('transform', False))
        return et


class Haude(ETBase):
    """
    only requires air temp and relative humidity at 2:00 pm. Good for moderate
    zones despite being simple

    References
    ----------
    [1]. Haude, W. (1954). Zur praktischen Bestimmung der aktuellen und potentiellen Evaporation und Evapotranspiration.
     Schweinfurter Dr. und Verlag-Ges..
    """
    def __call__(self, *args, **kwargs):

        etp = None  # f_mon * (6.11 × 10(7.48 × T / (237+T)) - rf × es)
        self.post_process(etp, kwargs.get('transform', False))
        return etp


class JensenHaiseBasins(ETBase):
    """
    This method generates daily pan evaporation (inches) using a coefficient for the month `cts`, , the daily
    average air temperature (F), a coefficient `ctx`, and solar radiation (langleys/day) as givn in
    BASINS program[2].
    The computations are
    based on the Jensen and Haise (1963) formula.
              PET = CTS * (TAVF - CTX) * RIN

        where
              PET = daily potential evapotranspiration (in)
              CTS = monthly variable coefficient
             TAVF = mean daily air temperature (F), computed from max-min
              CTX = coefficient
              RIN = daily solar radiation expressed in inches of evaporation

              RIN = SWRD/(597.3 - (.57 * TAVC)) * 2.54

        where
             SWRD = daily solar radiation (langleys)
             TAVC = mean daily air temperature (C)
    :uses cts float or array like. Value of monthly coefficient `cts` to be used. If float, then same value is
            assumed for all months. If array like then it must be of length 12.
    :uses ctx `float` constant coefficient value of `ctx` to be used in Jensen and Haise formulation.
    """
    def __call__(self, *args, **kwargs):
        if 'cts_jh' in kwargs:
            cts = kwargs['cts_jh']
        else:
            cts = self.cons['cts_jh']

        if 'cts_jh' in kwargs:
            ctx = kwargs['ctx_jh']
        else:
            ctx = self.cons['ctx_jh']

        if not isinstance(cts, float):
            if not isinstance(np.array(ctx), np.ndarray):
                raise ValueError('cts must be array like')
            else:  # if cts is array like it must be given for 12 months of year, not more not less
                if len(np.array(cts)) > 12:
                    raise ValueError('cts must be of length 12')
        else:  # if only one value is given for all moths distribute it as monthly value
            cts = np.array([cts for _ in range(12)])

        if not isinstance(ctx, float):
            raise ValueError('ctx must be float')

        # distributing cts values for all dates of input data
        self.input['cts'] = np.nan
        for m, i in zip(self.input.index.month, self.input.index):
            for _m in range(m):
                self.input.at[i, 'cts'] = cts[_m]

        cts = self.input['cts']
        taf = self.input['temp'].values

        rad_in = self.rad_to_evap()
        pan_evp = np.multiply(np.multiply(cts, np.subtract(taf, ctx)), rad_in)
        et = np.where(pan_evp < 0.0, 0.0, pan_evp)

        self.post_process(et, kwargs.get('transform', False))
        return et


class Kharrufa(ETBase):
    """
    For monthly potential evapotranspiration estimation, originally presented by Kharrufa, 1885. Xu and Singh, 2001
    presented following formula:
        et = 0.34 * p * Tmean**1.3
    et: pot. evapotranspiration in mm/month.
    Tmean: Average temperature in Degree Centigrade
    p: percentage of total daytime hours for the period used (daily or monthly) outof total daytime hours of the
       year (365 * 12)
    Kharrufa, N. S. (1985). Simplified equation for evapotranspiration in arid regions. Beitrage zur
	Hydrologie, 5(1), 39-47.
    """
    def __call__(self, *args, **kwargs):
        ta = self.input['temp']

        N = self.daylight_fao56()  # mean daily percentage of annual daytime hours
        n_annual = assign_yearly(N, self.input.index)

        et = 0.34 * n_annual['N'].values * ta**1.3

        self.post_process(et, kwargs.get('transform', False))
        return et


class Linacre(ETBase):
    """
     using formulation of Linacre 1977 who simplified Penman method.
     :uses
       temp
       tdew/rel_hum
       https://doi.org/10.1016/0002-1571(77)90007-3
     """
    def __call__(self, *args, **kwargs):

        if 'tdew' not in self.input:
            if 'rel_hum' in self.input:
                self.tdew_from_t_rel_hum()

        tm = np.add(self.input['temp'].values, np.multiply(0.006, self.cons['altitude']))
        tmp1 = np.multiply(500, np.divide(tm, 100 - self.cons['lat_dec_deg']))
        tmp2 = np.multiply(15, np.subtract(self.input['temp'].values, self.input['tdew'].values))
        upar = np.add(tmp1, tmp2)

        et = np.divide(upar, np.subtract(80, self.input['temp'].values))

        self.post_process(et, kwargs.get('transform', False))
        return et


class Makkink(ETBase):
    """
    :uses
      a_s, b_s
      temp
      solar_rad

    using formulation of Makkink
    """
    def __call__(self, *args, **kwargs):

        rs = self.rs()

        delta = self.slope_sat_vp(self.input['temp'].values)
        gamma = self.psy_const()

        et = np.subtract(np.multiply(np.multiply(0.61, np.divide(delta, np.add(delta, gamma))),
                                     np.divide(rs, 2.45)), 0.12)

        self.post_process(et, kwargs.get('transform', False))
        return et


class Irmak(ETBase):
    """
    Pandey et al 2016, presented 3 formulas for Irmak.
    1    eto = -0.611 + 0.149 * rs + 0.079 * t
    2    eto = -0.642 + 0.174 * rs + 0.0353 * t
    3    eto = -0.478 + 0.156 * rs - 0.0112 * tmax + 0.0733 * tmin

    References:
        Irmak 2003
        Tabari et al 2011
        Pandey et al 2016
    """


class Mahringer(ETBase):
    """
     Developed by Mahringer in Germany. [1] Wrote formula as
           eto = 0.15072 * sqrt(3.6) * (es - ea)
    """


class Mather(ETBase):
    """
    Developed by Mather 1978 and used in Rosenberry et al 2004. Calculates daily Pot ETP.
    pet = [1.6 (10T_a/I) ** 6.75e-7 * I**3 - 7.71e-7 * I**2 + 1.79e-2 * I + 0.49] (10/d)
    I = annual heat index, sum(Ta/5)1.514
    d = number of days in month
    """


class MattShuttleworth(ETBase):
    """
    using formulation of Matt-Shuttleworth and Wallace, 2009. This is designed for semi-arid and windy areas as an
    alternative to FAO-56 Reference Crop method
    10.13031/2013.29217
    https://elibrary.asabe.org/abstract.asp?aid=29217
    """
    def __call__(self, *args, **kwargs):

        self.requirements(constants=['CH', 'Roua', 'Ca', 'surf_res'])

        ch = self.cons['CH']  # crop height
        ro_a = self.cons['Roua']
        ca = self.cons['Ca']  # specific heat of the air
        # surface resistance (s m-1) of a well-watered crop equivalent to the FAO crop coefficient
        r_s = self.cons['surf_res']

        vabar = self.avp_from_rel_hum()  # Vapour pressure
        vas = self.mean_sat_vp_fao56()
        r_n = self.net_rad(vabar)  # net radiation
        u2 = self._wind_2m()    # Wind speed
        delta = self.slope_sat_vp(self.input['temp'].values)   # slope of vapour pressure curve
        gam = self.psy_const()    # psychrometric constant

        tmp1 = self.seconds * ro_a * ca
        # clinmatological resistance (s*m^-1) (S5.34)
        r_clim = np.multiply(tmp1, np.divide(np.subtract(vas, vabar), np.multiply(delta, r_n)))
        r_clim = np.where(r_clim == 0, 0.1, r_clim)   # correction for r_clim = 0
        u2 = np.where(u2 == 0, 0.1, u2)               # correction for u2 = 0

        #  ratio of vapour pressure deficits at 50m to vapour pressure deficits at 2m heights, eq S5.35
        a1 = (302 * (delta + gam) + 70 * gam * u2)
        a2 = (208 * (delta + gam) + 70 * gam * u2)
        a3 = 1/r_clim * ((302 * (delta + gam) + 70 * gam * u2) / (208 * (delta + gam) + 70 * gam * u2) * (208 / u2) - (302 / u2))
        vpd50_to_vpd2 = a1/a2 + a3

        # aerodynamic coefficient for crop height (s*m^-1) (eq S5.36 in McMohan et al 2013)
        a1 = 1 / (0.41**2)
        a2 = np.log((50 - 0.67 * ch) / (0.123 * ch))
        a3 = np.log((50 - 0.67 * ch) / (0.0123 * ch))
        a4 = np.log((2 - 0.08) / 0.0148) / np.log((50 - 0.08) / 0.0148)
        rc_50 = a1 * a2 * a3 * a4

        a1 = 1/LAMBDA
        a2 = (delta * r_n + (ro_a * ca * u2 * (vas - vabar)) / rc_50 * vpd50_to_vpd2)
        a3 = (delta + gam * (1 + r_s * u2 / rc_50))
        et = a1 * a2/a3
        self.post_process(et, kwargs.get('transform', False))
        return et


class McGuinnessBordne(ETBase):
    """
    calculates evapotranspiration [mm/day] using Mcguinnes Bordne formulation McGuinnes and Bordne, 1972.

    """
    def __call__(self, *args, **kwargs):

        ra = self._et_rad()
        # latent heat of vaporisation, MJ/Kg
        _lambda = LAMBDA  # multiply((2.501 - 2.361e-3), self.input['temp'].values)
        tmp1 = np.multiply((1/_lambda), ra)
        tmp2 = np.divide(np.add(self.input['temp'].values, 5), 68)
        et = np.multiply(tmp1, tmp2)

        self.post_process(et, kwargs.get('transform', False))
        return et


class Penman(ETBase):
    """
    calculates pan evaporation from open water using formulation of Penman, 1948, as mentioned (as eq 12) in
    McMahon et al., 2012. If wind data is missing then equation 33 from Valiantzas, 2006 is used which does not require
    wind data.

    uses:  wind_f='pen48', a_s=0.23, b_s=0.5, albedo=0.23
           uz
           temp
           rs
           reh_hum

    :param `wind_f` str, if 'pen48 is used then formulation of [1] is used otherwise formulation of [3] requires
             wind_f to be 2.626.
    """
    # todo, gives -ve values sometimes

    def __call__(self, **kwargs):
        self.requirements(constants=['lat_dec_deg', 'altitude', 'wind_f', 'albedo'],
                          ts=['temp', 'rh_mean'])

        if self.cons['wind_f'] not in ['pen48', 'pen56']:
            raise ValueError('value of given wind_f is not allowed.')

        wind_method = 'macmohan'
        if 'wind_method' in kwargs:
            wind_method = kwargs['wind_method']

        if self.cons['wind_f'] == 'pen48':
            _a = 2.626
            _b = 0.09
        else:
            _a = 1.313
            _b = 0.06

        delta = self.slope_sat_vp(self.input['temp'].values)
        gamma = self.psy_const()

        rs = self.rs()

        vabar = self.avp_from_rel_hum()  # Vapour pressure  *ea*
        r_n = self.net_rad(vabar, rs)  # net radiation
        vas = self.mean_sat_vp_fao56()

        if 'wind_speed' in self.input.columns:
            if self.verbosity > 1:
                print("Wind data have been used for calculating the Penman evaporation.")
            u2 = self._wind_2m(method=wind_method)
            fau = _a + 1.381 * u2
            ea = np.multiply(fau, np.subtract(vas, vabar))

            tmp1 = np.divide(delta, np.add(delta, gamma))
            tmp2 = np.divide(r_n, LAMBDA)
            tmp3 = np.multiply(np.divide(gamma, np.add(delta, gamma)), ea)
            evap = np.add(np.multiply(tmp1, tmp2), tmp3)
        # if wind data is not available
        else:
            if self.verbosity > 1:
                print("Alternative calculation for Penman evaporation without wind data has been performed")

            ra = self._et_rad()
            tmp1 = np.multiply(np.multiply(0.047, rs), np.sqrt(np.add(self.input['temp'].values, 9.5)))
            tmp2 = np.multiply(np.power(np.divide(rs, ra), 2.0), 2.4)
            tmp3 = np.multiply(_b, np.add(self.input['temp'].values, 20))
            tmp4 = np.subtract(1, np.divide(self.input['rh_mean'].values, 100))
            tmp5 = np.multiply(tmp3, tmp4)
            evap = np.add(np.subtract(tmp1, tmp2), tmp5)

        self.post_process(evap, kwargs.get('transform', False))
        return evap


class PenPan(ETBase):
    """
    implementing the PenPan formulation for Class-A pan evaporation as given in Rotstayn et al., 2006

    """
    def __call__(self, **kwargs):

        self.requirements(constants=['lat_dec_deg', 'altitude', 'pen_ap', 'albedo', 'alphaA', 'pan_over_est',
                                     'pan_est'],
                          ts=['temp', 'wind_speed'])

        epan = self.evap_pan()

        et = epan

        if self.cons['pan_over_est']:
            if self.cons['pan_est'] == 'pot_et':
                et = np.multiply(np.divide(et, 1.078), self.cons['pan_coef'])
            else:
                et = np.divide(et, 1.078)

        self.post_process(et, kwargs.get('transform', False))
        return et


class PenmanMonteith(ETBase):
    """
    calculates reference evapotrnaspiration according to Penman-Monteith (Allen et al 1998) equation which is
    also recommended by FAO. The etp is calculated at the time step determined by the step size of input data.
    For hourly or sub-hourly calculation, equation 53 is used while for daily time step equation 6 is used.

    # Requirements
    Following timeseries data is used
      relative humidity
      temperature

    Following constants are used
     lm=None, a_s=0.25, b_s=0.5, albedo=0.23

    http://www.fao.org/3/X0490E/x0490e08.htm#chapter%204%20%20%20determination%20of%20eto
    """
    def __call__(self, *args, **kwargs):
        self.requirements(constants=['lat_dec_deg', 'altitude', 'albedo', 'a_s', 'b_s'],
                          ts=['temp', 'wind_speed', 'jday'])

        wind_2m = self._wind_2m()

        d = self.slope_sat_vp(self.input['temp'].values)
        g = self.psy_const()

        # Mean saturation vapour pressure
        if 'es' not in self.input:
            if self.freq_in_mins == 1440:
                es = self.mean_sat_vp_fao56()
            elif self.freq_in_mins == 60:
                es = self.sat_vp_fao56(self.input['temp'].values)
            elif self.freq_in_mins < 60:   # TODO should sub-hourly be same as hourly?
                es = self.sat_vp_fao56(self.input['temp'].values)
            else:
                raise NotImplementedError
        else:
            es = self.input['es']

        # actual vapour pressure
        ea = self.avp_from_rel_hum()
        if 'vp_def' not in self.input:
            vp_d = es - ea   # vapor pressure deficit
        else:
            vp_d = self.input['vp_def']

        rn = self.net_rad(ea)              # eq 40 in Fao
        _g = self.soil_heat_flux(rn)

        t1 = 0.408 * (d*(rn - _g))
        nechay = d + g*(1 + 0.34 * wind_2m)

        if self.freq_in_mins == 1440:
            t5 = t1 / nechay
            t6 = 900/(self.input['temp']+273) * wind_2m * vp_d * g / nechay
            pet = np.add(t5, t6)

        elif self.freq_in_mins < 1440:  # TODO should sub-hourly be same as hourly?
            t3 = np.multiply(np.divide(37, self.input['temp']+273.0), g)
            t4 = np.multiply(t3, np.multiply(wind_2m, vp_d))
            upar = t1 + t4
            pet = upar / nechay
        else:
            raise NotImplementedError("For frequency of {} minutes, {} method can not be implemented"
                                      .format(self.freq_in_mins, self.name))
        self.post_process(pet, kwargs.get('transform', False))
        return pet


class PriestleyTaylor(ETBase):
    """
    following formulation of Priestley & Taylor, 1972.
    uses: , a_s=0.23, b_s=0.5, alpha_pt=1.26, albedo=0.23
    :param `alpha_pt` Priestley-Taylor coefficient = 1.26 for Priestley-Taylor model (Priestley and Taylor, 1972)
    https://doi.org/10.1175/1520-0493(1972)100<0081:OTAOSH>2.3.CO;2
     """
    def __call__(self, *args, **kwargs):
        self.requirements(constants=['lat_dec_deg', 'altitude', 'alpha_pt', 'albedo'])

        delta = self.slope_sat_vp(self.input['temp'].values)
        gamma = self.psy_const()
        vabar = self.avp_from_rel_hum()    # *ea*
        r_n = self.net_rad(vabar)   # net radiation
        # vas = self.mean_sat_vp_fao56()

        tmp1 = np.divide(delta, np.add(delta, gamma))
        tmp2 = np.multiply(tmp1, np.divide(r_n, LAMBDA))
        tmp3 = np.subtract(tmp2, np.divide(self.soil_heat_flux(), LAMBDA))
        et = np.multiply(self.cons['alpha_pt'], tmp3)

        self.post_process(et, kwargs.get('transform', False))
        return et


class Romanenko(ETBase):

    def __call__(self, *args, **kwargs):
        self.requirements(constants=['lat_dec_deg', 'altitude', 'albedo'],
                          ts=['temp'])
        """
        using formulation of Romanenko
        uses:
          temp
          rel_hum
        There are two variants of it in Song et al 2017.
        https://www.scirp.org/(S(czeh2tfqyw2orz553k1w0r45))/reference/ReferencesPapers.aspx?ReferenceID=2151471
        """
        t = self.input['temp'].values
        vas = self.mean_sat_vp_fao56()
        vabar = self.avp_from_rel_hum()  # Vapour pressure  *ea*

        tmp1 = np.power(np.add(1, np.divide(t, 25)), 2)
        tmp2 = np.subtract(1, np.divide(vabar, vas))
        et = np.multiply(np.multiply(4.5, tmp1), tmp2)

        self.post_process(et, kwargs.get('transform', False))
        return et


class SzilagyiJozsa(ETBase):
    """
    using formulation of Azilagyi, 2007.
     https://doi.org/10.1029/2006GL028708
    """
    def __call__(self, *args, **kwargs):

        self.requirements(constants=['wind_f', 'alphaPT'])

        if self.cons['wind_f'] == 'pen48':
            _a = 2.626
            _b = 0.09
        else:
            _a = 1.313
            _b = 0.06
        alpha_pt = self.cons['alphaPT']  # Priestley Taylor constant

        delta = self.slope_sat_vp(self.input['temp'].values)
        gamma = self.psy_const()

        rs = self.rs()
        vabar = self.avp_from_rel_hum()  # Vapour pressure  *ea*
        r_n = self.net_rad(vabar)   # net radiation
        vas = self.mean_sat_vp_fao56()

        if 'uz' in self.input.columns:
            if self.verbosity > 1:
                print("Wind data have been used for calculating the Penman evaporation.")
            u2 = self._wind_2m()
            fau = _a + 1.381 * u2
            ea = np.multiply(fau, np.subtract(vas, vabar))

            tmp1 = np.divide(delta, np.add(delta, gamma))
            tmp2 = np.divide(r_n, LAMBDA)
            tmp3 = np.multiply(np.divide(gamma, np.add(delta, gamma)), ea)
            et_penman = np.add(np.multiply(tmp1, tmp2), tmp3)
        # if wind data is not available
        else:
            if self.verbosity > 1:
                print("Alternative calculation for Penman evaporation without wind data have been performed")

            ra = self._et_rad()
            tmp1 = np.multiply(np.multiply(0.047, rs), np.sqrt(np.add(self.input['temp'].values, 9.5)))
            tmp2 = np.multiply(np.power(np.divide(rs, ra), 2.0), 2.4)
            tmp3 = np.multiply(_b, np.add(self.input['temp'].values, 20))
            tmp4 = np.subtract(1, np.divide(self.input['rh_mean'].values, 100))
            tmp5 = np.multiply(tmp3, tmp4)
            et_penman = np.add(np.subtract(tmp1, tmp2), tmp5)

        # find equilibrium temperature T_e
        t_e = self.equil_temp(et_penman)

        delta_te = self.slope_sat_vp(t_e)   # slope of vapour pressure curve at T_e
        # Priestley-Taylor evapotranspiration at T_e
        et_pt_te = np.multiply(alpha_pt, np.multiply(np.divide(delta_te, np.add(delta_te, gamma)), np.divide(r_n, LAMBDA)))
        et = np.subtract(np.multiply(2, et_pt_te), et_penman)

        self.post_process(et, kwargs.get('transform', False))
        return et


class Thornthwait(ETBase):
    """calculates reference evapotrnaspiration according to empirical temperature based Thornthwaite
    (Thornthwaite 1948) method. The method actualy calculates both ETP and evaporation. It requires only temperature
    and day length as input. Suitable for monthly values.
    """
    def __call__(self, *args, **kwargs):

        if 'daylight_hrs' not in self.input.columns:
            day_hrs = self.daylight_fao56()
        else:
            day_hrs = self.input['daylight_hrs']

        self.input['adj_t'] = np.where(self.input['temp'].values < 0.0, 0.0, self.input['temp'].values)
        I = self.input['adj_t'].resample('A').apply(custom_resampler)  # heat index (I)
        a = (6.75e-07 * I ** 3) - (7.71e-05 * I ** 2) + (1.792e-02 * I) + 0.49239
        self.input['a'] = a
        a_mon = self.input['a']    # monthly values filled with NaN
        a_mon = pd.DataFrame(a_mon)
        a_ann = pd.DataFrame(a)
        a_monthly = a_mon.merge(a_ann, left_index=True, right_index=True, how='left').fillna(method='bfill')
        self.input['I'] = I
        i_mon = self.input['I']  # monthly values filled with NaN
        i_mon = pd.DataFrame(i_mon)
        i_ann = pd.DataFrame(I)
        i_monthly = i_mon.merge(i_ann, left_index=True, right_index=True, how='left').fillna(method='bfill')

        tmp1 = np.multiply(1.6, np.divide(day_hrs, 12.0))
        tmp2 = np.divide(self.input.index.daysinmonth, 30.0)
        tmp3 = np.multiply(np.power(np.multiply(10.0, np.divide(self.input['temp'].values, i_monthly['I'].values)),
                                    a_monthly['a'].values), 10.0)
        pet = np.multiply(tmp1, np.multiply(tmp2, tmp3))

        # self.input['Thornthwait_daily'] = np.divide(self.input['Thornthwait_Monthly'].values, self.input.index.days_in_month)
        self.post_process(pet, kwargs.get('transform', False))
        return pet


class MortonCRAE(ETBase):
    """
    for monthly pot. ET and wet-environment areal ET and actual ET by Morton 1983.
    :return:

    """


class Papadakis(ETBase):
    """
    Calculates monthly values based on saturation vapor pressure and temperature. Following equation is given by
        eto = 0.5625 * (ea_tmax - ed)
        ea: water pressure corresponding to avg max temperature [KiloPascal].
        ed: saturation water pressure corresponding to the dew point temperature [KiloPascal].
    Rosenberry et al., 2004 presented following equation quoting McGuinnes and Bordne, 1972
        pet = 0.5625 * [es_max - (es_min - 2)] (10/d)
        d = number of days in month
        es = saturated vapour pressure at temperature of air in millibars
    """


class Ritchie(ETBase):
    """
    Given by Jones and Ritchie 1990 and quoted by Valipour, 2005 and Pandey et al., 2016
      et = rs * alpha [0.002322 * tmax + 0.001548*tmin + 0.11223]
    """
    def __call__(self, *args, **kwargs):

        self.requirements(constants=['ritchie_a', 'ritchie_b', 'ritchie_b', 'ritchie_alpha'],
                          ts=['tmin', 'tmax'])
        ritchie_a = self.cons['ritchie_a']
        ritchie_b = self.cons['ritchie_b']
        ritchie_c = self.cons['ritchie_c']
        alpha = self.cons['ritchie_alpha']
        rs = self.rs()
        eto = rs * alpha * [ritchie_a * self.input['tmax'] + ritchie_b * self.input['tmin'] + ritchie_c]

        self.post_process(eto, kwargs.get('transform', False))
        return eto


class Turc(ETBase):
    """
    The original formulation is from Turc, 1961 which was developed for southern France and Africa.
    Pandey et al 2016 mentioned a modified version of Turc quoting Xu et al., 2008, Singh, 2008 and Chen and Chen, 2008.
                         eto = alpha_t * 0.013 T/(T+15) ( (23.8856Rs + 50)/gamma)

    A shorter version of this formula is quoted by Valipour, 2015 quoting Xu et al., 2008
                         eto = (0.3107 * Rs + 0.65) [T alpha_t / (T + 15)]

    Here it is implemented as given (as eq 5) in Alexandris, et al., 2008 which is;
                for rh > 50 %:
                   eto = 0.0133 * [T_mean / (T_mean + 15)] ( Rs + 50)
                for rh < 50 %:
                   eto = 0.0133 * [T_mean / (T_mean + 15)] ( Rs + 50) [1 + (50 - Rh) / 70]

    uses
    :param `k` float or array like, monthly crop coefficient. A single value means same crop coefficient for
          whole year
    :param `a_s` fraction of extraterrestrial radiation reaching earth on sunless days
    :param `b_s` difference between fracion of extraterrestrial radiation reaching full-sun days
             and that on sunless days.
    Turc, L. (1961). Estimation of irrigation water requirements, potential evapotranspiration: a simple climatic
    formula evolved up to date. Ann. Agron, 12(1), 13-49.
    """
    def __call__(self, *args, **kwargs):

        self.requirements(constants=['lat_dec_deg', 'altitude', 'turc_k'],
                          ts=['temp'])

        use_rh = False  # because while testing daily, rhmin and rhmax are given and rhmean is calculated by default
        if 'use_rh' in kwargs:
            use_rh = kwargs['use_rh']

        rs = self.rs()
        ta = self.input['temp'].values
        et = np.multiply(np.multiply(self.cons['turc_k'], (np.add(np.multiply(23.88, rs), 50))),
                         np.divide(ta, (np.add(ta, 15))))

        if use_rh:
            if 'rh_mean' in self.input.columns:
                rh_mean = self.input['rh_mean'].values
                eq1 = np.multiply(np.multiply(np.multiply(self.cons['turc_k'], (np.add(np.multiply(23.88, rs), 50))),
                                              np.divide(ta, (np.add(ta, 15)))),
                                  (np.add(1, np.divide((np.subtract(50, rh_mean)), 70))))
                eq2 = np.multiply(np.multiply(self.cons['turc_k'], (np.add(np.multiply(23.88, rs), 50))),
                                  np.divide(ta, (np.add(ta, 15))))
                et = np.where(rh_mean < 50, eq1, eq2)

        self.post_process(et, kwargs.get('transform', False))
        return et


class Valiantzas(ETBase):
    """
    Djaman 2016 mentioned 2 methods from him, however Valipour 2015 tested 5 variants of his formulations in Iran.
    Ahmad et al 2019 used 6 variants of this method however, Djaman et al., 2017 used 9 of its variants.
    These 9 methods are given below:
    method_1:
      This is equation equation 19 in Valiantzas, 2012. This also does not require wind data.
      eto = 0.0393 * Rs* sqrt(T_avg + 9.5) - (0.19 * Rs**0.6 * lat_rad**0.15)
           + 0.0061(T_avg + 20)(1.12*_avg - T_min - 2)**0.7

    method_2:
      This is equation 14 in Valiantzas, 2012. This does not require wind data. The recommended value of alpha is 0.23.
      eto = 0.0393 * Rs * sqrt(T_avg + 9.5) - (0.19 * Rs**0.6 * lat_rad**0.15) + 0.078(T_avg + 20)(1 - rh/100)

    method_3
      eto = 0.0393 * Rs * sqrt(T_avg + 9.5) - (Rs/Ra)**2 - [(T_avg + 20) * (1-rh/100) * ( 0.024 - 0.1 * Waero)]

    method_4:
      This is equation 35 in Valiantzas 2013c paper and was reffered as Fo-PENM method with using alpha as 0.25.
      eto = 0.051 * (1 - alpha) * Rs * sqrt(T_avg + 9.5) - 2.4 * (Rs/Ra)**2
          + [0.048 * (T_avg + 20) * ( 1- rh/100) * (0.5 + 0.536 * u2)] + (0.00012 * z)

    method_5:
      This is equation 30 in Valiantzas 2013c. This is when no wind data is available.
      eto = 0.0393 * Rs sqrt(T_avg + 9.5)
           - [2.46 * Rs * lat**0.15 / (4 * sin(2 * pi * J / 365 - 1.39) lat + 12)**2 + 0.92]**2
           - 0.024 * (T_avg + 20)(1 - rh/100) - (0.0268 * Rs)
           + (0.0984 * (T_avg + 17)) * (1.03 + 0.00055) * (T_max - T_min)**2 - rh/100

    method_6
      This method is when wind speed and solar radiation data is not available. This is equation 34 in
      Valiantzas, 2013c.
      eto = 0.0068 * Ra * sqrt[(T_avg + 9.5) * (T_max - T_min)]
          - 0.0696 * (T_max - T_min) - 0.024 * (T_avg + 20)
          * [ ((1-rh/100) - 0.00455 * Ra * sqrt(T_max - T_dew)
          + 0.0984 * (T_avg + 17) * (1.03 + 0.0055) * (T_max - T_min)**2)
          - rh/100

    method_7:
      This is equation 27 in Valiantzas, 2013c. This method requires all data. Djaman et al., (by mistake) used 0.0043
      in denominator instead of 0.00043.
      eto = 0.051 * (1-alpha) * Rs * (T_avg + 9.5)**0.5
            - 0.188 * (T_avg + 13) * (Rs/Ra - 0.194)
            * (1 - 0.00015) * (T_avg + 45)**2 * sqrt(rh/100)
            - 0.0165 * Rs * u**0.7 + 0.0585 * (T_avg + 17) * u**0.75
            * {[1 + 0.00043 * (T_max - T_min)**2]**2 - rh/100} / [1 + 0.00043 * (T_max - T_min)**2 + 0.0001*z]

    method_8:
      eto = 0.051 * (1-alpha) * Rs * (T_avg + 9.5)**0.5
            - 2.4 (Rs/Ra)**2 - 2.4 * (T_avg + 20) * (1 - rh/100)
            - 0.0165 * Rs * u**0.7 + 0.0585 * (T_avg + 17) * u**0.75
            * { [1 + 0.00043 (T_max - T_min)**2]**2 - rh/100} / ( 1 + 0.00043 * (T_max - T_min)**2 + (0.0001 * z)

    method_9:
      This must be equation 29 of Valiantzas, 2013c but method 9 in Djaman et al., 2017 used 2.46 instead of 22.46. This
      formulation does not require Ra.
      eto = [0.051 * (1-alpha) * Rs (T_avg + 9.5)**2
            * (2.46 * Rs * lat**0.15 / (4 sin(2 * pi J / 365 - 1.39) * lat + 12)**2 + 0.92)]**2
            - 0.024 * (T_avg + 20) * (1-rh/100) - 0.0165 * Rs * u**0.7
            + 0.0585 * (T_avg + 17) * u**0.75 * {[(1.03 + 0.00055) * (T_max - T_min)**2 - rh/100] + 0.0001*z}
    """
    def __call__(self, method='method_1', **kwargs):

        self.requirements(constants=['valiantzas_alpha'],
                          ts=['temp'])

        alpha = self.cons['valiantzas_alpha']
        z = self.cons['altitute']
        rh = self.input['rh']
        ta = self.input['temp']
        tmin = self.input['tmin']
        tmax = self.input['tmax']
        j = self.input['jday']
        ra = self._et_rad()
        u2 = self._wind_2m()
        w_aero = np.where(rh <= 65.0, 1.067, 0.78)  # empirical weighted factor
        rs_ra = (self.rs() / ra)**2
        tr = tmax - tmin
        tr_sq = tr**2
        lat_15 = self.lat_rad**0.15

        t_sqrt = np.sqrt(ta + 9.5)

        init = 0.0393 * self.rs() * t_sqrt

        rs_fact = 0.19 * (self.rs()**0.6) * lat_15

        t_20 = ta + 20

        rh_factor = 1.0 - (rh/100.0)

        if method == 'method_1':
            eto = init - rs_fact + 0.0061 * t_20 * (1.12 * (ta - tmin) - 2.0)**0.7

        elif method == 'method_2':
            eto = init - rs_fact + (0.078 * t_20 * rh_factor)

        elif method == 'method_3':
            eto = init - rs_ra - (t_20 * rh_factor * (0.024 - 0.1 * w_aero))

        elif method == 'method_4':
            eto = 0.051 * (1 - alpha) * self.rs() * t_sqrt - 2.4 * rs_ra + (0.048 * t_20 * rh_factor * (0.5 + 0.536 * u2)) + (0.00012 * z)

        elif method == 'method_5':
            eto = init

        elif method == 'method_6':
            pass

        elif method == 'method_7':
            pass

        elif method == 'method_8':
            pass

        elif method == 'method_9':
            eto = 0.051 * (1 - alpha) * self.rs() * (ta + 9.5)**2 * (2.46 * self.rs() * lat_15) / (4 * np.sin(2 * 3.14 * j / 365 - 1.39))

        else:
            raise ValueError

        self.post_process(eto, kwargs.get('transform', False))
        return eto

class Oudin(ETBase):
    """
    https://doi.org/10.1016/j.jhydrol.2004.08.026
    """
    pass

class RengerWessolek(ETBase):
    """
    RENGER, M. & WESSOLEK, G. (1990): Auswirkungen von Grundwasserabsenkung und Nutzungsänderungen auf die
     Grundwasserneubildung. – Mit. Inst. für Wasserwesen, Univ. der Bundeswehr München, 386: 295-307.
    """

class Black(ETBase):
    """
     https://doi.org/10.2136/sssaj1969.03615995003300050013x
    """

class McNaughtonBlack(ETBase):
    """
     https://doi.org/10.1029/WR009i006p01579
    """


def custom_resampler(array_like):
    """calculating heat index using monthly values of temperature."""
    return np.sum(np.power(np.divide(array_like, 5.0), 1.514))


def assign_yearly(data, index):
    # TODO for leap years or when first or final year is not complete, the results are not correct immitate
    # https://github.com/cran/Evapotranspiration/blob/master/R/Evapotranspiration.R#L1848
    """ assigns `step` summed data to whole data while keeping the length of data preserved."""
    n_ts = pd.DataFrame(data, index=index, columns=['N'])
    a = n_ts.resample('A').sum()  # annual sum
    ad = a.resample('D').backfill()  # annual sum backfilled
    # for case
    if len(ad) < 2:
        ad1 = pd.DataFrame(np.full(data.shape, np.nan), pd.date_range(n_ts.index[0], periods=len(data), freq='D'),
                           columns=['N'])
        ad1.loc[ad1.index[-1]] = ad.values
        ad2 = ad1.bfill()
        return ad2
    else:
        idx = pd.date_range(n_ts.index[0], ad.index[-1], freq="D")
        n_df_ful = pd.DataFrame(np.full(idx.shape, np.nan), index=idx, columns=['N'])
        n_df_ful['N'][ad.index] = ad.values.reshape(-1, )
        n_df_obj = n_df_ful[n_ts.index[0]: n_ts.index[-1]]
        n_df_obj1 = n_df_obj.bfill()
        return n_df_obj1



__all__ = ['TemporalFusionTransformer']

"""
This file contain most of the code from
https://github.com/google-research/google-research/blob/master/tft/libs/tft_model.py
The TemporalFusionTransformer class has been modified so that it can be used as regular layer and without static,
categorical, observation and future inputs. It has also been modified to use 1D CNN as
an alternative of LSTM
The original code had Apache Licence, Version 2.0 although a lot of
the code has been modified here. http://www.apache.org/licenses/LICENSE-2.0
"""

import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.layers import TimeDistributed, Dense, InputLayer, Embedding, Conv1D, Flatten

from .utils import concatenate, gated_residual_network, add_and_norm, apply_gating_layer, get_decoder_mask
from .utils import InterpretableMultiHeadAttention


# Layer definitions.
stack = tf.keras.backend.stack


class TemporalFusionTransformer(tf.keras.layers.Layer):
    """
    Implements the model of https://arxiv.org/pdf/1912.09363.pdf
    This layer applies variable selection three times. First on static inputs,
    then on encoder inputs and then on decoder inputs. The corresponding weights
    are called `static_weights`, `historical_weights` and `future_weights` respectively.

    1, 11, 21, 31, a
    2, 12, 22, 32, b
    3, 13, 23, 33, c
    4, 14, 24, 34, d

    Parameters
    ---------
        hidden_units : int
            determines the depth/weight matrices size in TemporalFusionTransformer.
        num_encoder_steps : int
            lookback steps used in the model.
        num_heads : int
            must be>=1, number of attention heads to be used in MultiheadAttention layer.
        num_inputs : int
            number of input features
        total_time_steps : int
            greater than num_encoder_steps, This is sum of lookback steps + forecast length.
            Forecast length is the number of horizons to be predicted.
        known_categorical_inputs : list
            a,b,c
        input_obs_loc :
        unknown_inputs :
        static_inputs :
        static_input_loc None/list:
            location of static inputs
        category_counts : list
            Number of categories per categorical variable
        use_cnn : bool
            whether to use cnn or not. If not, then lstm will be used otherwise
            1D CNN will be used with "causal" padding.
        kernel_size : int
            kernel size for 1D CNN. Only valid if use_cnn is True.
        use_cudnn : bool
            default False, Whether to use Keras CuDNNLSTM or standard LSTM layers
        dropout_rate : float
            default 0.1, >=0 and <=1 amount of dropout to be used at GRNs.
        future_inputs : bool
            whether the given data contains futre known observations or not.
        return_attention_components bool:
            If True, then this layer (upon its call) will return outputs + attention
            componnets. Attention components are dictionary consisting of following keys
            and their values as numpy arrays.

        return_sequences bool:
            if True, then output and attention weights will consist of  encoder_lengths/lookback
            and decoder_length/forecast_len. Otherwise predictions for only decoder_length will be
            returned.

    Example
    -------
        >>> params = {'num_inputs': 3, 'total_time_steps': 192, 'known_regular_inputs': [0, 1, 2]}
        >>> output_size = 1
        >>> quantiles = [0.25, 0.5, 0.75]
        >>> layers = {
        >>>   "Input": {"config": {"shape": (params['total_time_steps'], params['num_inputs']), 'name': "Model_Input"}},
        >>>   "TemporalFusionTransformer": {"config": params},
        >>>   "lambda": {"config": tf.keras.layers.Lambda(lambda _x: _x[Ellipsis, -1, :])},
        >>>   "Dense": {"config": {"units": output_size * len(quantiles)}},
        >>>   'Reshape': {'target_shape': (3, 1)}}

    """
    def __init__(
            self,
            hidden_units: int,
            num_encoder_steps: int,
            num_heads: int,
            num_inputs: int,
            total_time_steps: int,
            known_categorical_inputs,
            static_input_loc,
            category_counts,
            known_regular_inputs,
            input_obs_loc,
            use_cnn: bool = False,
            kernel_size: int = None,
            # stack_size:int = 1,
            use_cudnn: bool = False,
            dropout_rate: float = 0.1,
            future_inputs: bool = False,
            return_attention_components: bool = False,
            return_sequences: bool = False,
            **kwargs
    ):

        if use_cudnn:
            assert kernel_size is not None

        self.time_steps = total_time_steps
        self.input_size = num_inputs
        self.use_cnn = use_cnn
        self.kernel_size = kernel_size
        self._known_regular_input_idx = known_regular_inputs  # [1,2,3]
        self._input_obs_loc = input_obs_loc  # [0]
        self._static_input_loc = static_input_loc  # [3,4]
        self.category_counts = category_counts  # [2, 2]
        self._known_categorical_input_idx = known_categorical_inputs

        # Network params
        self.use_cudnn = use_cudnn  # Whether to use GPU optimised LSTM
        self.hidden_units = int(hidden_units)
        self.dropout_rate = float(dropout_rate)
        self.encoder_steps = num_encoder_steps  # historical steps/lookback steps
        self.num_heads = int(num_heads)
        # self.num_stacks= int(stack_size) # todo

        self.future_inputs = future_inputs
        self.return_attention_components = return_attention_components
        self.return_sequences = return_sequences

        super().__init__(**kwargs)

    def __call__(self, alle_inputs, *args, **kwargs):
        """Returns graph defining layers of the TFT.
        """

        # Size definitions.
        encoder_steps = self.encoder_steps  # encoder_steps

        unknown_inputs, known_combined_layer, obs_inputs, static_inputs = self.get_tft_embeddings(alle_inputs)

        # known_combined_layer.shape = (num_examples, time_steps, hidden_units, num_outputs)
        # obs_inputs.shape = (num_examples, time_steps, hidden_units, 1)
        # static_inputs.shape = (num_examples, num_cat_variables, hidden_units)

        # Isolate known and observed historical inputs.
        if unknown_inputs is not None:
            historical_inputs = concatenate([
                unknown_inputs[:, :encoder_steps, :],
                known_combined_layer[:, :encoder_steps, :],
                obs_inputs[:, :encoder_steps, :]
            ], axis=-1, name="historical_inputs")
        else:
            if obs_inputs is not None:
                # we are extracting only historical data i.e. lookback from obs_inputs
                # (num_examples, encoder_steps, hidden_units, 4) <- (num_examples, encoder_steps, hidden_units, num_outputs) | (num_examples, encoder_steps, hidden_units, 1)
                historical_inputs = concatenate([
                    known_combined_layer[:, :encoder_steps, :],
                    obs_inputs[:, :encoder_steps, :]], axis=-1, name="historical_inputs")
            else:
                historical_inputs = known_combined_layer[:, :encoder_steps, :]

        if self.future_inputs:
            assert self.time_steps - self.encoder_steps > 0
            # Isolate only known future inputs.
            future_inputs = known_combined_layer[:, encoder_steps:, :]  # (num_examples, 24, hidden_units, num_outputs)
        else:
            assert self.time_steps == self.encoder_steps
            future_inputs = None

        def static_combine_and_mask(embedding):
            """Applies variable selection network to static inputs.

            Args:
              embedding: Transformed static inputs (num_examples, num_cat_variables, hidden_units)

            Returns:
              Tensor output for variable selection network
            """

            # Add temporal features
            _, num_static, _ = embedding.get_shape().as_list()

            flatten = tf.keras.layers.Flatten()(embedding)  # (num_examples, hidden_units*num_cat_variables)

            # Nonlinear transformation with gated residual network.
            mlp_outputs = gated_residual_network(  # (num_examples, num_cat_variables)
                flatten,
                self.hidden_units,
                output_size=num_static,
                dropout_rate=self.dropout_rate,
                use_time_distributed=False,
                additional_context=None,
                name='GRN_static'
            )

            # (num_examples, num_cat_variables)
            sparse_weights = tf.keras.layers.Activation('softmax', name='sparse_static_weights')(mlp_outputs)
            sparse_weights = K.expand_dims(sparse_weights, axis=-1)  # (num_examples, num_cat_variables, 1)

            trans_emb_list = []
            for i in range(num_static):
                e = gated_residual_network(  # e.shape = (num_examples, 1, hidden_units)
                    embedding[:, i:i + 1, :],
                    self.hidden_units,
                    dropout_rate=self.dropout_rate,
                    use_time_distributed=False,
                    name=f'GRN_static_{i}'
                )
                trans_emb_list.append(e)

            # (num_examples, num_cat_variables, hidden_units)
            transformed_embedding = concatenate(trans_emb_list, axis=1, name="transfomred_embedds")

            # (num_examples, num_cat_variables, hidden_units)
            combined = tf.keras.layers.Multiply(name="StaticWStaticEmb")(
                [sparse_weights, transformed_embedding])

            static_vec = K.sum(combined, axis=1)  # (num_examples, hidden_units)

            return static_vec, sparse_weights

        static_context_state_h = None
        static_context_state_c = None
        static_context_variable_selection = None
        static_context_enrichment = None
        static_weights = None

        if static_inputs is not None:
            static_encoder, static_weights = static_combine_and_mask(static_inputs)

            # static_encoder.shape = (num_examples, hidden_units)
            # static_weights.shape = (num_examples, num_cat_variables, 1)
            static_context_variable_selection = gated_residual_network(  # (num_examples, hidden_units)
                static_encoder,
                self.hidden_units,
                dropout_rate=self.dropout_rate,
                use_time_distributed=False,
                name="GNR_st_cntxt_var_select"
            )
            static_context_enrichment = gated_residual_network(  # (num_examples, hidden_units)
                static_encoder,
                self.hidden_units,
                dropout_rate=self.dropout_rate,
                use_time_distributed=False,
                name="GRN_st_cntxt_enrich"
            )
            static_context_state_h = gated_residual_network(  # (num_examples, hidden_units)
                static_encoder,
                self.hidden_units,
                dropout_rate=self.dropout_rate,
                use_time_distributed=False,
                name="GRN_st_cntxt_h"
            )
            static_context_state_c = gated_residual_network(  # (num_examples, hidden_units)
                static_encoder,
                self.hidden_units,
                dropout_rate=self.dropout_rate,
                use_time_distributed=False,
                name="GRN_st_cntxt_c"
            )

        def lstm_combine_and_mask(embedding, static_context, _name=None):
            """Apply temporal variable selection networks.

            Args:
              embedding: Transformed inputs. (num_examples, time_steps, hidden_units, num_inputs)
              # time_steps can be either encoder_steps or decoder_steps.
              # num_inputs can be either encoder_inputs or decoder_inputs
              static_context:
              _name: name of encompassing layers

            Returns:
              Processed tensor outputs. (num_examples, time_steps, hidden_units)
            """

            # Add temporal features
            _, time_steps, embedding_dim, num_inputs = embedding.get_shape().as_list()

            flatten = K.reshape(embedding,  # (num_examples, time_steps, num_inputs*hidden_units)
                                [-1, time_steps, embedding_dim * num_inputs])

            if static_context is not None:
                _expanded_static_context = K.expand_dims(  # (num_examples, 1, hidden_units)
                    static_context, axis=1)
            else:
                _expanded_static_context = None

            # Variable selection weights
            mlp_outputs, static_gate = gated_residual_network(
                flatten,
                self.hidden_units,
                output_size=num_inputs,
                dropout_rate=self.dropout_rate,
                use_time_distributed=False,
                additional_context=_expanded_static_context,
                return_gate=True,
                name=f'GRN_with_{_name}'
            )

            # mlp_outputs.shape (num_examples, time_steps, num_inputs)
            # static_gate.shape (num_examples, time_steps, num_inputs)
            # sparse_weights (num_examples, time_steps, num_inputs)
            sparse_weights = tf.keras.layers.Activation('softmax',    # --> (num_examples, time_steps, num_inputs)
                                                        name=f'sparse_{_name}_weights_softmax')(mlp_outputs)
            sparse_weights = tf.expand_dims(sparse_weights, axis=2)  # (num_examples, time_steps, 1, num_inputs)

            # Non-linear Processing & weight application
            trans_emb_list = []
            for i in range(num_inputs):
                grn_output = gated_residual_network(  # --> (num_examples, time_steps, hidden_units)
                    embedding[Ellipsis, i],   # feeding (num_examples, time_steps, hidden_units) as input
                    self.hidden_units,
                    dropout_rate=self.dropout_rate,
                    use_time_distributed=False,
                    name=f'GRN_with_{_name}_for_{i}'
                )
                trans_emb_list.append(grn_output)

            # (num_examples, time_steps, hidden_units, num_inputs)
            transformed_embedding = stack(trans_emb_list, axis=-1)

            # --> (num_examples, time_steps, hidden_units, num_inputs)
            combined = tf.keras.layers.Multiply(name=f'sparse_and_transform_{_name}')(
                [sparse_weights, transformed_embedding])
            temporal_ctx = K.sum(combined, axis=-1)  # (num_examples, time_steps, hidden_units)

            return temporal_ctx, sparse_weights, static_gate

        historical_features, historical_weights, historical_gate = lstm_combine_and_mask(
            historical_inputs,
            static_context_variable_selection,
            _name='history'
        )
        # historical_features.shape = (num_examples, encoder_steps, hidden_units)
        # historical_flags (num_examples, encoder_steps, 1, 4)
        future_features = None
        future_weights = None
        if future_inputs is not None:
            future_features, future_weights, future_gate = lstm_combine_and_mask(
                future_inputs,
                static_context_variable_selection,
                _name='future')

        # future_features = (num_examples, decoder_length, hidden_units)
        # future_flags = (num_examples, decoder_length, 1, num_outputs)

        initial_states = None
        if static_context_state_h is not None:
            initial_states = [static_context_state_h, static_context_state_c]

        if self.use_cnn:
            history_lstm = build_cnn(historical_features,
                                     self.hidden_units,
                                     self.kernel_size,
                                     return_state=True,
                                     _name="history",
                                     use_cudnn=self.use_cudnn)
        else:
            lstm = get_lstm(self.hidden_units, return_state=True, _name="history", use_cudnn=self.use_cudnn)
            history_lstm, state_h, state_c = lstm(historical_features, initial_state=initial_states)
        # history_lstm = (num_examples, encoder_steps, hidden_units)

        if future_features is not None:
            if self.use_cnn:
                future_lstm = build_cnn(future_features, self.hidden_units, self.kernel_size,
                                        return_state=False, _name="Future", use_cudnn=self.use_cudnn)

            else:
                lstm = get_lstm(self.hidden_units, return_state=False, _name="Future", use_cudnn=self.use_cudnn)
                future_lstm = lstm(future_features, initial_state=[state_h, state_c])

            # future_lstm = (num_examples, decoder_length, hidden_units)

            lstm_output = concatenate([history_lstm, future_lstm],
                                      axis=1,
                                      name='history_plus_future_lstm')  # (num_examples, time_steps, hidden_units)
            # Apply gated skip connection
            input_embeddings = concatenate([historical_features, future_features],
                                           axis=1,
                                           name="history_plus_future_embeddings"
                                           )  # (num_examples, time_steps, hidden_units)
        else:
            lstm_output = history_lstm
            input_embeddings = historical_features

        lstm_output, _ = apply_gating_layer(  # (num_examples, time_steps, hidden_units)
            lstm_output, self.hidden_units, self.dropout_rate, activation=None, name='GatingOnLSTM')
        # (num_examples, time_steps, hidden_units)
        temporal_feature_layer = add_and_norm([lstm_output, input_embeddings], name='AfterLSTM')

        # Static enrichment layers
        expanded_static_context = None
        if static_context_enrichment is not None:
            # (num_examples, 1, hidden_units)
            expanded_static_context = K.expand_dims(static_context_enrichment, axis=1)

        atten_input, _ = gated_residual_network(  # (num_examples, time_steps, hidden_units)
            temporal_feature_layer,
            self.hidden_units,
            dropout_rate=self.dropout_rate,
            use_time_distributed=False,
            additional_context=expanded_static_context,
            return_gate=True,
            name='GRN_temp_feature'
        )

        # Decoder self attention
        self_attn_layer = InterpretableMultiHeadAttention(
            self.num_heads, self.hidden_units, dropout=self.dropout_rate, name="InterpMultiHeadAtten")

        mask = get_decoder_mask(atten_input)
        # in some implementation cases, queries contain only decoder_length part but since official google repo
        # used all the inputs i.e. encoder+decoder part, we are doing so as well.
        # This is more useful in cases since transformer layer can be used as many to many.
        # Thus current behaviour is similar to `return_sequences=True` of LSTM.
        if self.return_sequences:
            queries = atten_input
        else:
            queries = atten_input[:, self.encoder_steps:]

        # queries (batch_size, time_steps, hidden_units
        # atten_input (batch_size, time_steps, hidden_units
        atten_output, self_att = self_attn_layer(queries,
                                                 atten_input, atten_input, mask=mask)

        # atten_output (batch_size, time_steps, hidden_units)
        # self_att (num_heads, batch_size, time_steps, time_steps
        # x =  (num_examples, time_steps, hidden_units)
        atten_output, _ = apply_gating_layer(  # # x =  (num_examples, time_steps, hidden_units)
            atten_output,
            self.hidden_units,
            dropout_rate=self.dropout_rate,
            activation=None,
            name="GatingOnX"
        )
        # # x =  (num_examples, time_steps, hidden_units)
        atten_output = add_and_norm([atten_output, queries], name="XAndEnriched")

        # Nonlinear processing on outputs
        decoder = gated_residual_network(  # # x =  (num_examples, time_steps, hidden_units)
            atten_output,
            self.hidden_units,
            dropout_rate=self.dropout_rate,
            use_time_distributed=False,
            name="NonLinearityOnOut"
        )

        # Final skip connection
        decoder, _ = apply_gating_layer(decoder, self.hidden_units, activation=None,
                                        name="FinalSkip")  # # x =  (num_examples, time_steps, hidden_units)

        # (num_examples, time_steps, hidden_units)
        transformer_output = add_and_norm([decoder, temporal_feature_layer], name="DecoderAndTempFeature")

        # Attention components for explainability
        attention_components = {
            # Temporal attention weights
            'decoder_self_attn': self_att,  # (num_atten_heads, num_examples, time_steps, time_steps)
            # Static variable selection weights  # (num_examples, 1)
            'static_variable_selection_weights': static_weights[Ellipsis, 0] if static_weights is not None else None,
            # Variable selection weights of past inputs  # (num_examples, encoder_steps, input_features)
            'encoder_variable_selection_weights': historical_weights[Ellipsis, 0, :],
            # Variable selection weights of future inputs
            # (num_examples, decoder_steps, input_features)
            'decoder_variable_selection_weights': future_weights[Ellipsis, 0, :] if future_weights is not None else None
        }

        self.attention_components = attention_components

        if self.return_attention_components:
            return transformer_output, attention_components

        return transformer_output

    def get_tft_embeddings(self, all_inputs):
        """Transforms raw inputs to embeddings.

        Applies linear transformation onto continuous variables and uses embeddings
        for categorical variables.

        Args:
          all_inputs: Inputs to transform [batch_size, time_steps, input_features]
          whre time_steps include both lookback and forecast. The input_features dimention of all_inputs can
          contain following inputs. `static_inputs`, `obs_inputs`, `categorical_inputs`, `regular_inputs`.

        Returns:
          Tensors for transformed inputs.
          unknown_inputs:
          known_combined_layer: contains regular inputs and categorical inputs (all known)
          obs_inputs: target values to be used as inputs.
          static_inputs
        """

        time_steps = self.time_steps

        # Sanity checks
        for i in self._known_regular_input_idx:
            if i in self._input_obs_loc:
                raise ValueError('Observation cannot be known a priori!')
        if self._input_obs_loc is not None:
            for i in self._input_obs_loc:
                if i in self._static_input_loc:
                    raise ValueError('Observation cannot be static!')

        if all_inputs.get_shape().as_list()[-1] != self.input_size:
            raise ValueError(
                'Illegal number of inputs! Inputs observed={}, expected={}'.format(
                  all_inputs.get_shape().as_list()[-1], self.input_size))

        num_categorical_variables = len(self.category_counts)  # 1
        num_regular_variables = self.input_size - num_categorical_variables  # 4

        embeddings = []
        if num_categorical_variables > 0:
            embedding_sizes = [   # [hidden_units]
                self.hidden_units for i, size in enumerate(self.category_counts)
            ]

            for i in range(num_categorical_variables):

                embedding = tf.keras.Sequential([InputLayer([time_steps]),Embedding(
                    self.category_counts[i],
                    embedding_sizes[i],
                    input_length=time_steps,
                    dtype=tf.float32)
                                                 ])
                embeddings.append(embedding)

            # regular_inputs
            regular_inputs, categorical_inputs = all_inputs[:, :, :num_regular_variables], \
                                                 all_inputs[:, :, num_regular_variables:]

            # regular_inputs (num_examples, time_steps, 4)
            # categorical_inputs (num_examples, time_steps, num_static_inputs)
            # list of lengnth=(num_static_inputs) with shape (num_examples, time_steps, hidden_units)
            embedded_inputs = [
                embeddings[i](categorical_inputs[Ellipsis, i])
                for i in range(num_categorical_variables)
            ]

        else:
            embedded_inputs = []
            # --> (num_examples, total_time_steps, num_inputs)
            regular_inputs = all_inputs[:, :, :num_regular_variables]
            categorical_inputs = None

        # Static inputs
        if len(self._static_input_loc) > 0:  # [4]
            static_inputs = [tf.keras.layers.Dense(self.hidden_units, name=f'StaticInputs{i}')(
                regular_inputs[:, 0, i:i + 1]) for i in range(num_regular_variables)
                           if i in self._static_input_loc] + [embedded_inputs[i][:, 0, :]
                   for i in range(num_categorical_variables)
                   if i + num_regular_variables in self._static_input_loc]

            # (num_examples, num_cat_variables, hidden_units) <-  [(num_examples, hidden_units)]
            static_inputs = stack(static_inputs, axis=1)

        else:  # there are not static inputs
            static_inputs = None

        def convert_real_to_embedding(_x, _name=None):
            """Applies linear transformation for time-varying inputs."""
            return TimeDistributed(Dense(self.hidden_units, name=_name))(_x)

        # whether we have and want to use target observations as inputs or not?
        if len(self._input_obs_loc) > 0:
            # Targets
            obs_inputs = stack([          # (num_examples, time_steps, hidden_units, 1)
                convert_real_to_embedding(regular_inputs[Ellipsis, i:i + 1], _name='InputObsDense')
                for i in self._input_obs_loc], axis=-1)
        else:
            obs_inputs = None

        wired_embeddings = []
        if num_categorical_variables > 0:
            # Observed (a prioir unknown) inputs
            for i in range(num_categorical_variables):
                if i not in self._known_categorical_input_idx \
                and i not in self._input_obs_loc:
                    e = embeddings[i](categorical_inputs[:, :, i])
                    wired_embeddings.append(e)

        unknown_inputs = []
        for i in range(regular_inputs.shape[-1]):
            if i not in self._known_regular_input_idx \
              and i not in self._input_obs_loc:
                e = convert_real_to_embedding(regular_inputs[Ellipsis, i:i + 1], _name=f'RegularInputsDense{i}')
                unknown_inputs.append(e)

        if unknown_inputs + wired_embeddings:
            if len(wired_embeddings) > 0:
              unknown_inputs = stack(unknown_inputs + wired_embeddings, axis=-1)
        else:
          unknown_inputs = None

        # A priori known inputs
        known_regular_inputs = [  # list of tensors all of shape (num_examples, total_time_steps, hidden_units)
            # feeding (num_examples, total_time_steps, 1) at each loop
            convert_real_to_embedding(regular_inputs[Ellipsis, i:i + 1], _name=f'KnownRegularInputs')
            for i in self._known_regular_input_idx
            if i not in self._static_input_loc
        ]
        known_categorical_inputs = [         # []
            embedded_inputs[i]
            for i in self._known_categorical_input_idx
            if i + num_regular_variables not in self._static_input_loc
        ]

        # (num_examples, time_steps, hidden_units, num_outputs)
        known_combined_layer = stack(known_regular_inputs + known_categorical_inputs, axis=-1)

        return unknown_inputs, known_combined_layer, obs_inputs, static_inputs


def build_cnn(inputs, filters, kernel_size, return_state, _name, use_cudnn=True):

    cnn = Conv1D(filters=filters, kernel_size=kernel_size, padding="causal", name=_name)
    cnn_output = cnn(inputs)

    return cnn_output


# LSTM layer
def get_lstm(hidden_units, return_state, _name=None, use_cudnn=True):
    """Returns LSTM cell initialized with default parameters."""
    if use_cudnn:
        lstm = tf.keras.layers.CuDNNLSTM(
            hidden_units,
            return_sequences=True,
            return_state=return_state,
            stateful=False,
            name=_name
        )
    else:
        lstm = tf.keras.layers.LSTM(
            hidden_units,
            return_sequences=True,
            return_state=return_state,
            stateful=False,
            # Additional params to ensure LSTM matches CuDNN, See TF 2.0 :
            # (https://www.tensorflow.org/api_docs/python/tf/keras/layers/LSTM)
            activation='tanh',
            recurrent_activation='sigmoid',
            recurrent_dropout=0,
            unroll=False,
            use_bias=True,
            name=_name
        )
    return lstm
